import logging
logger = logging.getLogger('isitfit')

from isitfit.utils import mergeSeriesOnTimestampRange
import numpy as np


class Ec2Common:
    def _handle_ec2obj(self, context_ec2):
        # parse out
        ec2_obj = context_ec2['ec2_obj']

        # logger.debug("%s, %s"%(ec2_obj.instance_id, ec2_obj.instance_type))

        # pandas series of CPU utilization, daily max, for 90 days
        df_metrics = context_ec2['df_metrics']

        # pandas series of number of cpu's available on the machine over time, past 90 days
        df_type_ts1 = context_ec2['df_type_ts1']

        # this is redundant with the implementation in _cloudwatch_metrics_core,
        # and it's here just in case the cached redis version is not a date,
        # but it's not really worth it to make a full refresh of the cache for this
        # if df_metrics.Timestamp.dtype==dt.date:
        try:
          df_metrics['Timestamp'] = df_metrics.Timestamp.dt.date
        except AttributeError:
          pass

        # convert type timeseries to the same timeframes as pcpu and n5mn
        #if ec2_obj.instance_id=='i-069a7808addd143c7':
        ec2_df = mergeSeriesOnTimestampRange(df_metrics, df_type_ts1)
        #logger.debug("\nafter merge series on timestamp range")
        #logger.debug(ec2_df.head())

        # merge with type changes (can't use .merge on timestamps, need to use .concat)
        #ec2_df = df_metrics.merge(df_type_ts2, left_on='Timestamp', right_on='EventTime', how='left')
        # ec2_df = pd.concat([df_metrics, df_type_ts2], axis=1)

        # merge with catalog
        ec2_df = ec2_df.merge(context_ec2['df_cat'][['API Name', 'cost_hourly']], left_on='instanceType', right_on='API Name', how='left')
        #logger.debug("\nafter merge with catalog")
        #logger.debug(ec2_df.head())

        # calculate number of running hours
        # In the latest 90 days, sampling is per minute in cloudwatch
        # https://aws.amazon.com/cloudwatch/faqs/
        # Q: What is the minimum resolution for the data that Amazon CloudWatch receives and aggregates?
        # A: ... For example, if you request for 1-minute data for a day from 10 days ago, you will receive the 1440 data points ...
        ec2_df['nhours'] = np.ceil(ec2_df.SampleCount/60)

        # set in context
        context_ec2['ec2_df'] = ec2_df

        # done
        return context_ec2


    def after_all(self, context_all):
        # unpack
        ec2_noCloudwatch, ec2_noCloudtrail = context_all['ec2_noCloudwatch'], context_all['ec2_noCloudtrail']


        # get now + 10 minutes
        # http://stackoverflow.com/questions/6205442/ddg#6205529
        import datetime as dt
        dt_now = dt.datetime.now()
        TRY_IN = 10
        now_plus_10 = dt_now + dt.timedelta(minutes = TRY_IN)
        now_plus_10 = now_plus_10.strftime("%H:%M")

        if len(ec2_noCloudwatch)>0:
          n_no_cw = len(ec2_noCloudwatch)
          has_more_cw = "..." if n_no_cw>5 else ""
          l_no_cw = ", ".join(ec2_noCloudwatch[:5])
          logger.warning("No cloudwatch data for %i resources: %s%s"%(n_no_cw, l_no_cw, has_more_cw))
          logger.warning("Try again in %i minutes (at %s) to check for new data"%(TRY_IN, now_plus_10))
          logger.info("")

        if len(ec2_noCloudtrail)>0:
          n_no_ct = len(ec2_noCloudtrail)
          has_more_ct = "..." if n_no_ct>5 else ""
          l_no_ct = ", ".join(ec2_noCloudtrail[:5])
          logger.warning("No cloudtrail data for %i resources: %s%s"%(n_no_ct, l_no_ct, has_more_ct))
          logger.warning("Try again in %i minutes (at %s) to check for new data"%(TRY_IN, now_plus_10))
          logger.info("")

        return context_all



import logging
logger = logging.getLogger('isitfit')

import pandas as pd
from tabulate import tabulate

# https://pypi.org/project/termcolor/
from termcolor import colored



class UtilizationListener:

  def __init__(self, ctx):
    # iterate over all ec2 instances
    self.sum_capacity = 0
    self.sum_used = 0
    self.df_all = []
    self.table = None # will contain the final table after calling `after_all`
    self.ctx = ctx


  def per_ec2(self, context_ec2):
    """
    Listener function to be called upon the download of each EC2 instance's data
    ec2_obj - boto3 resource
    ec2_df - pandas dataframe with data from cloudwatch+cloudtrail
    mm - mainManager class
    ddg_df - dataframe of data from datadog: {cpu,ram}-{max,avg}
    """
    # parse out context keys
    ec2_obj, ec2_df, mm, ddg_df = context_ec2['ec2_obj'], context_ec2['ec2_df'], context_ec2['mainManager'], context_ec2['ddg_df']

    # results: 2 numbers: capacity (USD), used (USD)
    res_capacity = (ec2_df.nhours*ec2_df.cost_hourly).sum()

    if 'ram_used_avg.datadog' in ec2_df.columns:
      # use both the CPU Average from cloudwatch and the RAM average from datadog
      utilization_factor = ec2_df[['Average', 'ram_used_avg.datadog']].mean(axis=1, skipna=True)
    else:
      # use only the CPU average from cloudwatch
      utilization_factor = ec2_df.Average

    res_used     = (ec2_df.nhours*ec2_df.cost_hourly*utilization_factor/100).sum()
    #logger.debug("res_capacity=%s, res_used=%s"%(res_capacity, res_used))

    self.sum_capacity += res_capacity
    self.sum_used += res_used
    self.df_all.append({'instance_id': ec2_obj.instance_id, 'capacity': res_capacity, 'used': res_used})


  def after_all(self, context_all):
    # for debugging
    df_all = pd.DataFrame(self.df_all)
    logger.debug("\ncapacity/used per instance")
    logger.debug(df_all)
    logger.debug("\n")

    return context_all



def tagsContain(f_tn, ec2_obj):
  if ec2_obj.tags is None:
    return False
  
  for t in ec2_obj.tags:
    for k in ['Key', 'Value']:
      if f_tn in t[k].lower():
        return True
  
  return False



class Ec2TagFilter:
  def __init__(self, filter_tags):
    self.filter_tags = filter_tags

  def per_ec2(self, context_ec2):
    # if filters requested, check that this instance passes

    # set in context
    context_ec2['filter_tags'] = self.filter_tags

    if self.filter_tags is None:
      # to continue with other listeners
      return context_ec2

    f_tn = self.filter_tags.lower()
    passesFilter = tagsContain(f_tn, context_ec2['ec2_obj'])
    if not passesFilter:
      # break other listeners
      return None

    # otherwise continue
    return context_ec2



import logging
logger = logging.getLogger('isitfit')

from .tagsSuggestBasic import TagsSuggestBasic
from ..utils import MAX_ROWS
import os
import json
from ..apiMan import ApiMan

class TagsSuggestAdvanced(TagsSuggestBasic):

  def __init__(self, ctx):
    logger.debug("TagsSuggestAdvanced::constructor")

    # api manager
    self.api_man = ApiMan(tryAgainIn=2, ctx=ctx)

    # proceed with parent constructor
    return super().__init__(ctx)


  def prepare(self):
    logger.debug("TagsSuggestAdvanced::prepare")
    self.api_man.register()


  def suggest(self):
    logger.info("Uploading ec2 names to s3")

    logger.debug("TagsSuggestAdvanced::suggest")
     
    # if status is not ok yet, ping again
    if self.api_man.r_register['isitfitapi_status']['code']=='Registration in progress':
        self.api_man.register()

    # boto3 s3 client
    s3_client  = self.api_man.boto3_session.client('s3' )

    import tempfile
    with tempfile.NamedTemporaryFile(suffix='.csv', prefix='isitfit-ec2names-', delete=True) as fh:
      logger.debug("Will use temporary file %s"%fh.name)
      self.tags_df.to_csv(fh.name, index=False)
      self.s3_key_suffix = 'tags_request.csv'
      s3_path = os.path.join(self.api_man.r_sts['Account'], self.api_man.r_sts['UserId'], self.s3_key_suffix)

      logger.debug("s3 PUT bucket=%s path=%s"%(self.api_man.r_body['s3_bucketName'], s3_path))
      s3_client.put_object(Bucket=self.api_man.r_body['s3_bucketName'], Key=s3_path, Body=fh)

    # POST /tags/suggest
    r2, dt_now = self._tags_suggest()
    if 'Message' in r2:
        from ..utils import IsitfitCliError
        raise IsitfitCliError(r2['Message'], self.ctx)

    # now listen on sqs
    any_found = False
    for m in self.api_man.listen_sqs('tags suggest', dt_now):
      # if done
      if m is None: break

      # process messages
      any_found = True
      logger.info("Server message: %s"%m.body_decoded['status'])
      if m.body_decoded['status'] != 'calculation complete':
        continue

      if m.body_decoded['status'] == 'calculation complete':
        # upon calculation complete message
        if 's3_key_suffix' not in m.body_decoded:
          logger.debug("(Missing s3_key_suffix key from body. Aborting)")
          return

        self.csv_fn = None
        with tempfile.NamedTemporaryFile(suffix='.csv', prefix='isitfit-tags-suggestAdvanced-', delete=False) as fh:
          self.csv_fn = fh.name
          s3_path = os.path.join(self.api_man.r_body['s3_keyPrefix'], m.body_decoded['s3_key_suffix'])
          logger.info("Downloading tag suggestions from isitfit server")
          logger.debug("Getting s3 file %s"%s3_path)
          logger.debug("Saving it into %s"%fh.name)
          response = s3_client.get_object(Bucket=self.api_man.r_body['s3_bucketName'], Key=s3_path)
          fh.write(response['Body'].read())

        logger.debug("TagsSuggestAdvanced:suggest .. read_csv")
        import pandas as pd
        self.suggested_df = pd.read_csv(self.csv_fn, nrows=MAX_ROWS)

        # count number of rows in csv
        # https://stackoverflow.com/a/36973958/4126114
        logger.debug("TagsSuggestAdvanced:suggest .. count_rows")
        with open(fh.name) as f2:
            self.suggested_shape = [sum(1 for line in f2), 4] # 4 is just hardcoded number of columns that doesn't matter much

        logger.debug("TagsSuggestAdvanced:suggest .. done")
        return

    # if nothing returned on sqs
    if not any_found:
      logger.error("Absolute radio silence on sqs :(")

    # either no sqs messages,
    # or found some sqs messages, but none were for tags request fulfilled
    import pandas as pd
    self.suggested_df = pd.DataFrame()
    self.suggested_shape = [0,4]
    self.csv_fn = None


  def _tags_suggest(self):
      logger.info("Requesting tag suggestions from isitfit server")

      load_send = {}
      #load_send.update(self.api_man.r_sts)
      load_send['s3_key_suffix'] = self.s3_key_suffix
      load_send['sqs_url'] = self.api_man.r_body['sqs_url']

      # curl equivalent
      # curl -X POST --data "foo=bar" https://api.isitfit.io/v0/974668457921/AIDA6F3WEM7AXY6Y4VWDC/tags/suggest
      r2, dt_now = self.api_man.request(
        method='post',
        relative_url='./tags/suggest',
        payload_json=load_send,
        authenticated_user_path=True
      )

      return r2, dt_now



from isitfit.cost.redshift.analyzer import AnalyzerBase, AnalyzerAnalyze, AnalyzerOptimize
import datetime as dt
import pandas as pd
import pytz
dt_now_d = dt.datetime.utcnow().replace(tzinfo=pytz.utc)


def test_redshiftPricingDict():
  from isitfit.cost.redshift.analyzer import redshiftPricing_dict
  assert len(redshiftPricing_dict.keys()) > 0



class TestAnalyzerBase:
  def test_init(self):
    ra = AnalyzerBase()
    assert True # no exception


class TestAnalyzerAnalyze:

  def test_fetch(self, mocker):
    mockreturn = lambda *args, **kwargs: pd.DataFrame({'Timestamp': [], 'Average': []})
    mockee = 'isitfit.cost.redshift.cloudwatchman.CloudwatchRedshift.handle_main'
    mocker.patch(mockee, side_effect=mockreturn)

    import datetime as dt
    import pytz
    dt_now_d = dt.datetime.utcnow().replace(tzinfo=pytz.utc)
    ex_iter = [
      ({'ClusterIdentifier': 'abc', 'NodeType': 'dc2.large', 'NumberOfNodes': 3, 'ClusterCreateTime': dt_now_d, 'Region': 'bla'},
        #pd.DataFrame([{'Average': 1, 'Timestamp': dt_now_d}]),
        'def',
        dt_now_d
      ),
    ]

    # prepare
    ra = AnalyzerAnalyze()
    ra.analyze_list = ex_iter

    # run and test
    ra.after_all({'click_ctx': None})
    assert ra.analyze_df.shape[0] == 1


  def test_calculate(self):
    ra = AnalyzerAnalyze()
    ra.analyze_df = pd.DataFrame([
      {'CostUsed': 1, 'CostBilled': 100, 'Region': 'bla'}
    ])
    ra.calculate({})
    assert ra.cwau_percent == 1


class TestAnalyzerOptimize:

  def test_fetch(self, mocker):
    mockreturn = lambda *args, **kwargs: pd.DataFrame({'Timestamp': [], 'Average': [], 'Maximum': [], 'Minimum': []})
    mockee = 'isitfit.cost.redshift.cloudwatchman.CloudwatchRedshift.handle_main'
    mocker.patch(mockee, side_effect=mockreturn)

    ex_iter = [
      ( {'ClusterIdentifier': 'def', 'NodeType': 'dc2.large', 'NumberOfNodes': 3, 'Region': 'bla'},
        # pd.DataFrame([{'Maximum': 1, 'Minimum': 1}]),
        'def',
        dt_now_d
      ),
    ]

    # prepare
    ra = AnalyzerOptimize()
    ra.analyze_list = ex_iter

    # run and test
    ra.after_all({'click_ctx': None})
    assert ra.analyze_df.shape[0] == 1


  def test_calculate(self):
    ra = AnalyzerOptimize()
    ra.analyze_df = pd.DataFrame([
      {'CpuMaxMax': 90, 'CpuMinMin': 80, 'Cost': 1, 'NumberOfNodes': 3},
      {'CpuMaxMax': 50, 'CpuMinMin':  1, 'Cost': 1, 'NumberOfNodes': 3},
    ])
    ra.calculate({})
    assert ra.analyze_df.classification.tolist() == ['Overused', 'Normal']



SECONDS_IN_ONE_DAY = 60*60*24 # 86400  # used for granularity (daily)


class NoCloudwatchException(Exception):
    pass


def raise_noCwExc(rc_id):
  raise NoCloudwatchException("No cloudwatch data for %s"%rc_id)


class NoCloudtrailException(Exception):
    pass


def mergeSeriesOnTimestampRange(df_cpu, df_type):
  """
  Upsamples df_type to df_cpu.
  Example:
    Input
      df_cpu = pd.Series({'time': [1,2,3,4], 'field_1': [5,6,7,8]})
      df_type = pd.Series({'time': [1,3], 'field_2': ['a','b']})
    Returns
      pd.Series({'time': [1,2,3,4], 'field_1': [5,6,7,8], 'field_2': ['a','a','b','b']})
  """
  import numpy as np

  df_cpu['instanceType'] = None
  # assume df_type is sorted in decreasing EventTime order (very important)
  # NB: since some instances are not present in the cloudtrail (for which we append artificially the "now" type)
  #     Need to traverse the df_type matrix backwards
  for index, row_type in df_type.iterrows():
      # use row_type.name instead of row_type['EventTime']
      # check note above about needing to traverse backwards
      # df_cpu.iloc[np.where(df_cpu.Timestamp >= row_type.name)[0], df_cpu.columns.get_loc('instanceType')] = row_type['instanceType']
      df_cpu.iloc[np.where(df_cpu.Timestamp <= row_type.name)[0], df_cpu.columns.get_loc('instanceType')] = row_type['instanceType']

  # fill na at beginning with back-fill
  # (artifact of cloudwatch having data at days before the creation of the instance)
  df_cpu['instanceType'] = df_cpu['instanceType'].fillna(method='backfill')
  return df_cpu







# copied from git-remote-aws
def mysetlocale():
  li = 'en_US.utf8'
  import os
  os.environ["LC_ALL"] = li
  os.environ["LANG"]   = li




MAX_ROWS = 10
MAX_COLS = 5
MAX_STRING = 20
def display_df(title, df, csv_fn, shape, logger):
    # https://pypi.org/project/termcolor/
    from termcolor import colored

    logger.info("")

    if shape[0]==0:
      logger.info(title)
      logger.info(colored("None", "red"))
      return

    if csv_fn is not None:
      logger.info(colored("The table '%s' was saved to the CSV file '%s'."%(title, csv_fn), "cyan"))
      logger.info(colored("It could be opened in the terminal with visidata (http://visidata.org/)","cyan"))
      logger.info(colored("and you can close visidata by pressing 'q'","cyan"))
      open_vd = input(colored('Would you like to do so? yes/[no] ', 'cyan'))
      if open_vd.lower() == 'yes' or open_vd.lower() == 'y':
        logger.info("Opening CSV file `%s` with visidata."%csv_fn)
        from subprocess import call
        call(["vd", csv_fn])
        logger.info("Exited visidata.")
        logger.info(colored("The table '%s' was saved to the CSV file '%s'."%(title, csv_fn), "cyan"))
        return
      else:
        logger.info("Not opening visidata.")
        logger.info("To open the results with visidata, use `vd %s`."%csv_fn)


    # if not requested to open with visidata
    from tabulate import tabulate
    df_show = df.head(n=MAX_ROWS)
    df_show = df_show.applymap(lambda c: (c[:MAX_STRING]+'...' if len(c)>=MAX_STRING else c) if type(c)==str else c)

    logger.info(tabulate(df_show, headers='keys', tablefmt='psql', showindex=False))

    if (shape[0] > MAX_ROWS) or (shape[1] > MAX_COLS):
      logger.info("...")
      logger.info("(results truncated)")
      # done
      return

    # done
    return


# Inherit from click's usageError since click can handle it automatically
# https://click.palletsprojects.com/en/7.x/exceptions/
from click import UsageError
class IsitfitCliError(UsageError):
  """
  Inherited from click.exceptions.UsageError
  because it adds the context as a constructor argument,
  which I need for checking "is_outdated"
  https://github.com/pallets/click/blob/8df9a6b2847b23de5c65dcb16f715a7691c60743/click/exceptions.py#L51
  """

  # exit code
  exit_code = 10

  # constructor parameters from
  # https://github.com/pallets/click/blob/8df9a6b2847b23de5c65dcb16f715a7691c60743/click/exceptions.py#L11
  def show(self, file=None):
    # ping matomo about error
    ping_matomo("/error?message=%s"%self.message)

    # continue
    from click._compat import get_text_stderr
    if file is None:
        file = get_text_stderr()

    # echo wrap
    color = 'red'
    import click
    def echo(message):
      # from click.utils import echo
      # echo('Error: %s' % self.format_message(), file=file, color=color)

      click.secho(message, fg=color)

    # main error
    echo('Error: %s' % self.format_message())

    # if isitfit installation is outdated, append a message to upgrade
    if self.ctx is not None:
      if self.ctx.obj.get('is_outdated', None):
        hint_1 = "Upgrade your isitfit installation with `pip3 install --upgrade isitfit` and try again."
        echo(hint_1)

    # add link to github issues
    hint_2 = "If the problem persists, please report it at https://github.com/autofitcloud/isitfit/issues/new"
    echo(hint_2)



def prompt_upgrade(pkg_name, current_version):
  """
  check if current version is out-of-date
  https://github.com/alexmojaki/outdated

  copied from https://github.com/WhatsApp/WADebug/blob/958ac37be804cc732ae514d4872b93d19d197a5c/wadebug/cli.py#L40
  """
  import outdated

  is_outdated = False
  try:
    is_outdated, latest_version = outdated.check_outdated(pkg_name, current_version)
  except ValueError as error:
    # catch case of "ValueError: Version 0.10.0 is greater than the latest version on PyPI: 0.9.1"
    # This would happen on my dev machine
    if not "is greater than" in str(error):
      raise

    # In this case, outdated does not cache the result to disk
    # so cache it myself (copied from https://github.com/alexmojaki/outdated/blob/565bb3fe1adc30da5e50249912cd2ac494662659/outdated/__init__.py#L61)
    latest_version = str(error).split(":")[1].strip()
    import datetime as dt
    import json
    with outdated.utils.cache_file(pkg_name, 'w') as f:
      try:
        data = [latest_version, outdated.utils.format_date(dt.datetime.now())]
        json.dump(data, f)
      except Exception as e:
        print('Error: ' + str(e))
        raise


  # is_outdated = True # FIXME for debugging
  if not is_outdated:
      return is_outdated

  import click
  msg_outdated = """The current version of {pkg_name} ({current_version}) is out of date.
Run `pip3 install {pkg_name} --upgrade` to upgrade to version {latest_version},
or use `isitfit --skip-check-upgrade ...` to skip checking for version upgrades of isitfit.
"""
  msg_outdated = msg_outdated.format(
      pkg_name=pkg_name, current_version=current_version, latest_version=latest_version
    )
  click.secho(msg_outdated, fg="red")

  # Give the user some time to read the message and possibly update
  import time
  from tqdm import tqdm
  wait_outdated = 10
  click.secho("Will continue in %i seconds"%wait_outdated, fg='yellow')
  for i in tqdm(range(wait_outdated)):
    time.sleep(1)

  return is_outdated


# This import needs to stay here for the sake of the mock in test_utils
import requests
SKIP_PING=False
def ping_matomo(action_name, uuid_val=None, isitfit_version=None):
  """
  Gather anonymous usage statistics
  """
  # get uuid
  from .dotMan import DotMan
  uuid_val = DotMan().get_myuid()

  # get version
  from . import isitfit_version as isitfit_cli_version

  # build action name field. note that "action_name" already starts with "/"
  full_actionName = "%s%s"%(isitfit_cli_version, action_name)

  # use base function
  from matomo_sdk_py.matomo_sdk_py import ping_matomo as ping_matomo_base
  ping_matomo_base(
    action_name=full_actionName,
    action_base="https://cli.isitfit.io",
    idsite=2, # 2 is for cli.isitfit.io
    uuid_val=uuid_val,
    matomo_url="https://isitfit.matomo.cloud/piwik.php"
  )


def display_footer():
    import logging
    logger = logging.getLogger('isitfit')

    from . import isitfit_version

    logger.info("")
    logger.info("⛅ Generated by isitfit version %s"%isitfit_version)
    logger.info("")
    logger.info("Useful links:")
    logger.info("ℹ️  isitfit homepage         https://isitfit.autofitcloud.com")
    logger.info("😞 isitfit issues           https://github.com/autofitcloud/isitfit/issues")
    logger.info("🌎 Global Climate Strike    https://twitter.com/hashtag/ClimateStrike")
    logger.info("❤️  Built by AutofitCloud    https://www.autofitcloud.com")



from click.core import Command as ClickCommand
class IsitfitCommand(ClickCommand):
    """
    Call display_footer at the end of each invokation
    https://github.com/pallets/click/blob/8df9a6b2847b23de5c65dcb16f715a7691c60743/click/core.py#L945
    """
    def invoke(self, *args, **kwargs):
        ret = super().invoke(*args, **kwargs)
        display_footer()
        return ret



def myreturn(df_xxx):
    if df_xxx.shape[0] > 0:
      return df_xxx
    else:
      return None # this means that the data was found in cache, but it was empty (meaning aws returned no data)





import logging
logger = logging.getLogger('isitfit')

import pandas as pd
from tabulate import tabulate

# https://pypi.org/project/termcolor/
from termcolor import colored

from isitfit.cost.redshift.reporter import ReporterBase

class ReporterAnalyzeEc2(ReporterBase):

  def postprocess(self, context_all):
    self.analyzer = context_all['analyzer']

    # unpack
    n_ec2_total, mm, n_ec2_analysed, region_include = context_all['n_ec2_total'], context_all['mainManager'], context_all['n_ec2_analysed'], context_all['region_include']
    self.analyzer.n_ec2_total = n_ec2_total
    self.analyzer.mm = mm
    self.analyzer.n_ec2_analysed = n_ec2_analysed
    self.analyzer.region_include = region_include

    # proceed
    cwau_val = 0
    if self.analyzer.sum_capacity!=0:
      cwau_val = self.analyzer.sum_used/self.analyzer.sum_capacity*100

    cwau_color = 'yellow'
    if cwau_val >= 70:
      cwau_color = 'green'
    elif cwau_val <= 30:
      cwau_color = 'red'

    dt_start = self.analyzer.mm.StartTime.strftime("%Y-%m-%d")
    dt_end   = self.analyzer.mm.EndTime.strftime("%Y-%m-%d")

    ri_max = 3
    ri_ell = "" if len(self.analyzer.region_include)<=ri_max else "..."
    ri_str = ", ".join(self.analyzer.region_include[:ri_max])+ri_ell
    
    self.table = [
            {'color': '',
             'label': "Start date",
             'value': "%s"%dt_start
            },
            {'color': '',
             'label': "End date",
             'value': "%s"%dt_end
            },
            {'color': '',
             'label': "Regions",
             'value': "%i (%s)"%(len(self.analyzer.region_include), ri_str)
            },
            {'color': '',
             'label': "EC2 machines (total)",
             'value': "%i"%self.analyzer.n_ec2_total
            },
            {'color': '',
             'label': "EC2 machines (analysed)",
             'value': "%i"%self.analyzer.n_ec2_analysed
            },
            {'color': 'cyan',
             'label': "Billed cost",
             'value': "%0.0f $"%self.analyzer.sum_capacity
            },
            {'color': 'cyan',
             'label': "Used cost",
             'value': "%0.0f $"%self.analyzer.sum_used
            },
            {'color': cwau_color,
             'label': "CWAU (Used/Billed)",
             'value': "%0.0f %%"%cwau_val
            },
    ]

    # done
    return context_all


  def display(self, context_all):
    def get_row(row):
        def get_cell(i):
          retc = row[i] if not row['color'] else colored(row[i], row['color'])
          return retc
        
        retr = [get_cell('label'), get_cell('value')]
        return retr

    dis_tab = [get_row(row) for row in self.table]

    # logger.info("Summary:")
    logger.info("Cost-Weighted Average Utilization (CWAU) of the AWS EC2 account:")
    logger.info("")
    logger.info(tabulate(dis_tab, headers=['Field', 'Value']))
    logger.info("")
    logger.info("For reference:")
    logger.info(colored("* CWAU >= 70% is well optimized", 'green'))
    logger.info(colored("* CWAU <= 30% is underused", 'red'))

    return context_all


  def email(self, context_all):
      """
      ctx - click context
      """
      context_2 = {}
      context_2['emailTo'] = context_all['emailTo']
      context_2['click_ctx'] = context_all['click_ctx']
      context_2['dataType'] = 'cost analyze' # ec2, not redshift
      context_2['dataVal'] = {'table': self.table}
      super().email(context_2)

      return context_all




class ReporterOptimizeEc2(ReporterBase):

  def __init__(self):
    self.csv_fn_final = None


  def postprocess(self, context_all):
    self.analyzer = context_all['analyzer']

    # unpack
    n_ec2_total, mm, n_ec2_analysed, region_include, df_cat = context_all['n_ec2_total'], context_all['mainManager'], context_all['n_ec2_analysed'], context_all['region_include'], context_all['df_cat']
    self.analyzer.n_ec2_total = n_ec2_total
    mm.df_cat = df_cat
    self.analyzer.mm = mm
    self.analyzer.n_ec2_analysed = n_ec2_analysed
    self.analyzer.region_include = region_include

    # process
    self._after_all()
    self._storecsv_all()
    return context_all

  def _after_all(self):
    df_all = pd.DataFrame(self.analyzer.ec2_classes)

    # if no data
    if df_all.shape[0]==0:
      self.df_sort = None
      self.sum_val = None
      return

    # merge current type hourly cost
    map_cost = self.analyzer.mm.df_cat[['API Name', 'cost_hourly']]
    df_all = df_all.merge(map_cost, left_on='instance_type', right_on='API Name', how='left').drop(['API Name'], axis=1)

    # merge the next-smaller instance type from the catalog for instances classified as Underused
    map_smaller = self.analyzer.mm.df_cat[['API Name', 'type_smaller', 'Linux On Demand cost_smaller']].rename(columns={'Linux On Demand cost_smaller': 'cost_hourly_smaller'})
    df_all = df_all.merge(map_smaller, left_on='instance_type', right_on='API Name', how='left').drop(['API Name'], axis=1)

    # merge next-larger instance type
    map_larger = self.analyzer.mm.df_cat[['API Name', 'type_smaller', 'cost_hourly']].rename(columns={'type_smaller': 'API Name', 'API Name': 'type_larger', 'cost_hourly': 'cost_hourly_larger'})
    df_all = df_all.merge(map_larger, left_on='instance_type', right_on='API Name', how='left').drop(['API Name'], axis=1)

    # convert from hourly to 3-months
    for fx1, fx2 in [('cost_3m', 'cost_hourly'), ('cost_3m_smaller', 'cost_hourly_smaller'), ('cost_3m_larger', 'cost_hourly_larger')]:
      df_all[fx1] = df_all[fx2] * 24 * 30 * 3
      df_all[fx1] = df_all[fx1].fillna(value=0).astype(int)

    # imply a recommended type
    from isitfit.cost.optimizerListener import class2recommendedType, class2recommendedCost
    df_all['recommended_type'] = df_all.apply(class2recommendedType, axis=1)
    df_all['savings'] = df_all.apply(class2recommendedCost, axis=1)
    df_all['savings'] = df_all.savings.fillna(value=0).astype(int)

    # keep a subset of columns
    df_all = df_all[['region', 'instance_id', 'instance_type', 'classification_1', 'classification_2', 'cost_3m', 'recommended_type', 'savings', 'tags']]

    # display
    #df_all = df_all.set_index('classification_1')
    #for v in ['Idle', 'Underused', 'Overused', 'Normal']:
    #  logger.info("\nInstance classification_1: %s"%v)
    #  if v not in df_all.index:
    #    logger.info("None")
    #  else:
    #    logger.info(df_all.loc[[v]]) # use double brackets to maintain single-row dataframes https://stackoverflow.com/a/45990057/4126114
    #
    #  logger.info("\n")

    # main results
    self.df_sort = df_all.sort_values(['savings'], ascending=True)
    self.sum_val = df_all.savings.sum()


  def _storecsv_all(self, *args, **kwargs):
      if self.df_sort is None:
        return

      import tempfile
      with tempfile.NamedTemporaryFile(prefix='isitfit-full-ec2-', suffix='.csv', delete=False) as  csv_fh_final:
        self.csv_fn_final = csv_fh_final.name
        logger.debug(colored("Saving final results to %s"%csv_fh_final.name, "cyan"))
        self.df_sort.to_csv(csv_fh_final.name, index=False)
        logger.debug(colored("Save complete", "cyan"))


  def display(self, context_all):
    if self.df_sort is None:
      logger.info(colored("No EC2 instances found", "red"))
      return context_all

    # display
    # Edit 2019-09-25 just show the full list. Will add filtering later. This way it's less ambiguous when all instances are "Normal"
    # self.df_sort.dropna(subset=['recommended_type'], inplace=True)
    
    # if no recommendations
    if self.df_sort.shape[0]==0:
      logger.info(colored("No optimizations from isitfit for this AWS EC2 account", "red"))
      return context_all
    
    # if there are recommendations, show them
    sum_comment = "extra cost" if self.sum_val>0 else "savings"
    sum_color = "red" if self.sum_val>0 else "green"

    #logger.info("Optimization based on the following CPU thresholds:")
    #logger.info(self.thresholds)
    #logger.info("")
    logger.info(colored("Recommended %s: %0.0f $ (over the next 3 months)"%(sum_comment, self.sum_val), sum_color))
    logger.info("")

    # display dataframe
    from isitfit.utils import display_df
    display_df(
      "Recommended EC2 size changes",
      self.df_sort,
      self.csv_fn_final,
      self.df_sort.shape,
      logger
    )

#    with pd.option_context("display.max_columns", 10):
#      logger.info("Details")
#      if self.df_sort.shape[0]<=10:
#        logger.info(df2tabulate(self.df_sort))
#      else:
#        logger.info(df2tabulate(self.df_sort.head(n=5)))
#        logger.info("...")
#        logger.info(df2tabulate(self.df_sort.tail(n=5)))
#        logger.info("")
#        logger.info(colored("Table originally with %i rows is truncated for top and bottom 5 only."%self.df_sort.shape[0], "cyan"))
#        logger.info(colored("Consider filtering it with --n=x for the 1st x results or --filter-tags=foo using a value from your own EC2 tags.", "cyan"))

    if self.analyzer.n!=0:
      logger.info(colored("This table has been filtered for only the 1st %i underused results"%self.analyzer.n, "cyan"))

    return context_all



import logging
logger = logging.getLogger('isitfit')


def dump_df_to_csv(df_dump, csv_prefix):
    import tempfile
    import pandas as pd

    # https://pypi.org/project/termcolor/
    from termcolor import colored

    with tempfile.NamedTemporaryFile(prefix=csv_prefix, suffix='.csv', delete=False) as fh:
      logger.info(colored("Dumping data into %s"%fh.name, "cyan"))
      df_dump.to_csv(fh.name, index=False)
      return fh.name



class TagsSuggestBasic:

  def __init__(self, ctx):
    logger.debug("TagsSuggestBasic::constructor")
    # boto3 ec2 and cloudwatch data
    import boto3
    self.ec2_resource = boto3.resource('ec2')
    self.tags_list = []
    self.tags_df = None
    self.ctx = ctx

  def prepare(self):
    logger.debug("TagsSuggestBasic::prepare")
    pass

  def tags_to_dict(self, ec2_obj):
    tags_dict = {x['Key']: x['Value'] for x in ec2_obj.tags if x['Key']=='Name'}
    return tags_dict

  def fetch(self):
    logger.debug("TagsSuggestBasic::fetch")
    logger.info("Counting EC2 instances")
    n_ec2_total = len(list(self.ec2_resource.instances.all()))
    msg_total = "Found a total of %i EC2 instances"%n_ec2_total
    if n_ec2_total==0:
      from ..utils import IsitfitCliError
      raise IsitfitCliError(msg_total, self.ctx)

    logger.warning(msg_total)

    self.tags_list = []
    from tqdm import tqdm
    desc = "Scanning EC2 instances"
    ec2_all = self.ec2_resource.instances.all()
    for ec2_obj in tqdm(ec2_all, total=n_ec2_total, desc=desc, initial=1):
      if ec2_obj.tags is None:
        tags_dict = {}
      else:
        tags_dict = self.tags_to_dict(ec2_obj)

      tags_dict['instance_id'] = ec2_obj.instance_id
      self.tags_list.append(tags_dict)

    # convert to pandas dataframe when done
    self.tags_df = self._list_to_df()


  def _list_to_df(self):
      logger.info("Converting tags list into dataframe")
      import pandas as pd
      df = pd.DataFrame(self.tags_list)
      df = df.rename(columns={'instance_id': '_0_instance_id', 'Name': '_1_Name'}) # trick to keep instance ID and name as the first columns
      df = df.sort_index(axis=1)  # sort columns
      df = df.rename(columns={'_0_instance_id': 'instance_id', '_1_Name': 'Name'}) # undo trick
      return df


  def suggest(self):
      logger.debug("TagsSuggestBasic::suggest")
      logger.info("Generating suggested tags")
      from .tagsImplier import TagsImplierMain
      tags_implier = TagsImplierMain(self.tags_df)
      self.suggested_df = tags_implier.imply()
      self.csv_fn = dump_df_to_csv(self.suggested_df, 'isitfit-tags-suggestBasic-')
      self.suggested_shape = self.suggested_df.shape


  def display(self):
    logger.debug("TagsSuggestBasic::display")
    from ..utils import display_df
    display_df(
      "Suggested tags:",
      self.suggested_df,
      self.csv_fn,
      self.suggested_shape,
      logger
    )



# Related
# https://docs.datadoghq.com/integrations/amazon_redshift/
# https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/redshift.html#Redshift.Paginator.DescribeClusters

from termcolor import colored

import logging
logger = logging.getLogger('isitfit')


class ReporterBase:
  def postprocess(self, context_all):
    raise Exception("To be implemented by derived class")

  def display(self, context_all):
    raise Exception("To be implemented by derived class")

  def email(self, context_all):
      """
      ctx - click context
      """
      for fx in ['dataType', 'dataVal']:
        if not fx in context_all:
          raise Exception("Missing field from context: %s. This function should be implemented by the derived class"%fx)

      # unpack
      emailTo, ctx = context_all['emailTo'], context_all['click_ctx']

      # check if email requested
      if emailTo is None:
          return context_all

      if len(emailTo)==0:
          return context_all

      from isitfit.emailMan import EmailMan
      em = EmailMan(
        dataType=context_all['dataType'], # ec2, not redshift
        dataVal=context_all['dataVal'],
        ctx=ctx
      )
      em.send(emailTo)

      return context_all



class ReporterAnalyze(ReporterBase):
  def postprocess(self, context_all):
    # unpack
    self.analyzer = context_all['analyzer']
    mm = context_all['mainManager']

    # copied from isitfit.cost.utilizationListener.after_all
    cwau_val = self.analyzer.cwau_percent
    cwau_color = 'yellow'
    if cwau_val >= 70: cwau_color = 'green'
    elif cwau_val <= 30: cwau_color = 'red'

    dt_start = mm.StartTime.strftime("%Y-%m-%d")
    dt_end   = mm.EndTime.strftime("%Y-%m-%d")

    self.table = [
      { 'color': '',
        'label': "Start date",
        'value': "%s"%dt_start
      },
      { 'color': '',
        'label': "End date",
        'value': "%s"%dt_end
      },
      { 'color': '',
        'label': "Regions",
        'value': "%i"%self.analyzer.regions_n
      },
      { 'color': '',
        'label': "Redshift clusters (total)",
        'value': "%i"%self.analyzer.n_rc_total
      },
      { 'color': '',
        'label': "Redshift clusters (analysed)",
        'value': "%i"%self.analyzer.n_rc_analysed
      },
      { 'color': '',
        'label': "Billed cost",
        'value': "%0.0f $"%self.analyzer.cost_billed
      },
      { 'color': '',
        'label': "Used cost",
        'value': "%0.0f $"%self.analyzer.cost_used
      },
      { 'color': cwau_color,
        'label': "CWAU",
        'value': "%0.0f %%"%cwau_val
      },
    ]
    return context_all


  def display(self, context_all):
    # copied from isitfit.cost.utilizationListener.display_all

    def get_row(row):
        def get_cell(i):
          retc = row[i] if not row['color'] else colored(row[i], row['color'])
          return retc

        retr = [get_cell('label'), get_cell('value')]
        return retr

    dis_tab = [get_row(row) for row in self.table]

    from tabulate import tabulate

    # logger.info("Summary:")
    logger.info("Cost-Weighted Average Utilization (CWAU) of the AWS Redshift account:")
    logger.info("")
    logger.info(tabulate(dis_tab, headers=['Field', 'Value']))
    logger.info("")
    logger.info("For reference:")
    logger.info(colored("* CWAU >= 70% is well optimized", 'green'))
    logger.info(colored("* CWAU <= 30% is underused", 'red'))
    logger.info("")
    logger.info("For the EC2 analysis, scroll up to the previous table.")
    return context_all


  def email(self, context_all):
      context_2 = {}
      context_2['emailTo'] = context_all['emailTo']
      context_2['click_ctx'] = context_all['click_ctx']
      context_2['dataType'] = 'cost analyze' # redshift, not ec2
      context_2['dataVal'] = {'table': self.table}
      super().email(context_2)

      return context_all




class ReporterOptimize(ReporterBase):
  def postprocess(self, context_all):
    # unpack
    self.analyzer = context_all['analyzer']

    # proceed
    analyze_df = self.analyzer.analyze_df
    analyze_df['CpuMaxMax'] = analyze_df['CpuMaxMax'].fillna(value=0).astype(int)
    analyze_df['CpuMinMin'] = analyze_df['CpuMinMin'].fillna(value=0).astype(int)

    # copied from isitfit.cost.optimizationListener.storecsv...
    import tempfile
    with tempfile.NamedTemporaryFile(prefix='isitfit-full-redshift-', suffix='.csv', delete=False) as  csv_fh_final:
      self.csv_fn_final = csv_fh_final.name
      logger.debug(colored("Saving final results to %s"%csv_fh_final.name, "cyan"))
      analyze_df.to_csv(csv_fh_final.name, index=False)
      logger.debug(colored("Save complete", "cyan"))

    return context_all


  def display(self, context_all):
    # copied from isitfit.cost.optimizationListener.display_all
    analyze_df = self.analyzer.analyze_df

    # display dataframe
    from ...utils import display_df
    display_df(
      "Redshift cluster classifications",
      analyze_df,
      self.csv_fn_final,
      analyze_df.shape,
      logger
    )
    return context_all


  def email(self, context_all):
      # silently return
      # raise Exception("Error emailing optimization: Not yet implemented")
      return context_all



import logging
logger = logging.getLogger('isitfit')

import click

# Use "cls" to use the IsitfitCommand class to show the footer
# https://github.com/pallets/click/blob/8df9a6b2847b23de5c65dcb16f715a7691c60743/click/decorators.py#L92
from ..utils import IsitfitCommand


@click.group(help="Evaluate AWS EC2 costs", invoke_without_command=False)
@click.option('--filter-region', default=None, help='specify a single region against which to run cost analysis/optimization')
@click.pass_context
def cost(ctx, filter_region):
  ctx.obj['filter_region'] = filter_region
  pass




@cost.command(help='Analyze AWS EC2 cost', cls=IsitfitCommand)
@click.option('--filter-tags', default=None, help='filter instances for only those carrying this value in the tag name or value')
@click.pass_context
def analyze(ctx, filter_tags):
    # gather anonymous usage statistics
    from ..utils import ping_matomo, IsitfitCliError
    ping_matomo("/cost/analyze")

    #logger.info("Is it fit?")
    logger.info("Initializing...")

    # moved these imports from outside the function to inside it so that `isitfit --version` wouldn't take 5 seconds due to the loading
    from ..cost.mainManager import MainManager
    from ..cost.cloudtrail_ec2type import CloudtrailCached
    from ..cost.utilizationListener import UtilizationListener
    from ..cost.cacheManager import RedisPandas as RedisPandasCacheManager
    from ..cost.datadogManager import DatadogCached
    from ..cost.ec2TagFilter import Ec2TagFilter
    from isitfit.cost.redshift.cloudwatchman import CloudwatchEc2
    from isitfit.cost.ec2.reporter import ReporterAnalyzeEc2
    from isitfit.ec2_catalog import Ec2Catalog
    from isitfit.cost.ec2.ec2Common import Ec2Common
    from isitfit.cost.redshift.iterator import Ec2Iterator


    share_email = ctx.obj.get('share_email', None)
    ul = UtilizationListener(ctx)

    # manager of redis-pandas caching
    cache_man = RedisPandasCacheManager()

    ddg = DatadogCached(cache_man)
    etf = Ec2TagFilter(filter_tags)
    cloudwatchman = CloudwatchEc2(cache_man)
    ra = ReporterAnalyzeEc2()
    mm = MainManager(ctx)
    ec2_cat = Ec2Catalog()
    ec2_common = Ec2Common()
    ec2_it = Ec2Iterator(filter_region=ctx.obj['filter_region'])

    # boto3 cloudtrail data
    cloudtrail_manager = CloudtrailCached(mm.EndTime, cache_man)

    # update dict and return it
    # https://stackoverflow.com/a/1453013/4126114
    inject_email_in_context = lambda context_all: dict({'emailTo': share_email}, **context_all)
    inject_analyzer = lambda context_all: dict({'analyzer': ul}, **context_all)

    # utilization listeners
    mm.set_iterator(ec2_it)
    mm.add_listener('pre', cache_man.handle_pre)
    mm.add_listener('pre', cloudtrail_manager.init_data)
    mm.add_listener('pre', ec2_cat.handle_pre)
    mm.add_listener('ec2', etf.per_ec2)
    mm.add_listener('ec2', cloudwatchman.per_ec2)
    mm.add_listener('ec2', cloudtrail_manager.single)
    mm.add_listener('ec2', ec2_common._handle_ec2obj)
    mm.add_listener('ec2', ddg.per_ec2)
    mm.add_listener('ec2', ul.per_ec2)
    mm.add_listener('all', ec2_common.after_all)
    mm.add_listener('all', ul.after_all)
    mm.add_listener('all', inject_analyzer)
    mm.add_listener('all', ra.postprocess)
    mm.add_listener('all', ra.display)
    mm.add_listener('all', inject_email_in_context)
    mm.add_listener('all', ra.email)

    # start download data and processing
    logger.info("Fetching history...")
    mm.get_ifi()

    # -----------------------------
    # now analyze redshift clusters
    logger.info("")
    logger.info("-"*20)
    from ..cost.redshift.cli import cost_analyze
    cost_analyze(share_email, filter_region=ctx.obj['filter_region'])
    logger.info("-"*20)



@cost.command(help='Generate recommendations of optimal EC2 sizes', cls=IsitfitCommand)
@click.option('--n', default=0, help='number of underused ec2 optimizations to find before stopping. Skip to get all optimizations')
@click.option('--filter-tags', default=None, help='filter instances for only those carrying this value in the tag name or value')
@click.pass_context
def optimize(ctx, n, filter_tags):
    # gather anonymous usage statistics
    from ..utils import ping_matomo, IsitfitCliError
    ping_matomo("/cost/optimize")

    #logger.info("Is it fit?")
    logger.info("Initializing...")

    # moved these imports from outside the function to inside it so that `isitfit --version` wouldn't take 5 seconds due to the loading
    from ..cost.mainManager import MainManager
    from ..cost.cloudtrail_ec2type import CloudtrailCached
    from ..cost.optimizerListener import OptimizerListener
    from ..cost.cacheManager import RedisPandas as RedisPandasCacheManager
    from ..cost.datadogManager import DatadogCached
    from ..cost.ec2TagFilter import Ec2TagFilter
    from isitfit.cost.redshift.cloudwatchman import CloudwatchEc2
    from isitfit.cost.ec2.reporter import ReporterOptimizeEc2
    from isitfit.ec2_catalog import Ec2Catalog
    from isitfit.cost.ec2.ec2Common import Ec2Common
    from isitfit.cost.redshift.iterator import Ec2Iterator

    ol = OptimizerListener(n)

    # manager of redis-pandas caching
    cache_man = RedisPandasCacheManager()

    ddg = DatadogCached(cache_man)
    etf = Ec2TagFilter(filter_tags)
    cloudwatchman = CloudwatchEc2(cache_man)
    ra = ReporterOptimizeEc2()
    mm = MainManager(ctx)
    ec2_cat = Ec2Catalog()
    ec2_common = Ec2Common()
    ec2_it = Ec2Iterator(filter_region=ctx.obj['filter_region'])

    # boto3 cloudtrail data
    cloudtrail_manager = CloudtrailCached(mm.EndTime, cache_man)

    # update dict and return it
    # https://stackoverflow.com/a/1453013/4126114
    inject_analyzer = lambda context_all: dict({'analyzer': ol}, **context_all)

    # utilization listeners
    mm.set_iterator(ec2_it)
    mm.add_listener('pre', cache_man.handle_pre)
    mm.add_listener('pre', cloudtrail_manager.init_data)
    mm.add_listener('pre', ol.handle_pre)
    mm.add_listener('pre', ec2_cat.handle_pre)
    mm.add_listener('ec2', etf.per_ec2)
    mm.add_listener('ec2', cloudwatchman.per_ec2)
    mm.add_listener('ec2', cloudtrail_manager.single)
    mm.add_listener('ec2', ec2_common._handle_ec2obj)
    mm.add_listener('ec2', ddg.per_ec2)
    mm.add_listener('ec2', ol.per_ec2)
    mm.add_listener('all', ec2_common.after_all)
    mm.add_listener('all', inject_analyzer)
    mm.add_listener('all', ra.postprocess)
    mm.add_listener('all', ra.display)


    # start download data and processing
    logger.info("Fetching history...")
    mm.get_ifi()

    # -----------------------------
    # now optimize redshift clusters
    logger.info("")
    logger.info("-"*20)
    from ..cost.redshift.cli import cost_optimize
    cost_optimize(filter_region=ctx.obj['filter_region'])
    logger.info("-"*20)




from .. import isitfit_version
import click

def version_core():
  print('isitfit version %s'%isitfit_version)

# Do not use the IsitfitCommand class to show the footer
# because the user will see the version info multiple times
@click.command(help="Show isitfit version")
def version():
  # This is redundant with isitfit --version (just in case a user calls "isitfit version")
  # The idea is similar to `git version` and `git --version`
  version_core()
  return





# RuntimeError: Click will abort further execution because Python 3 was configured to use ASCII as encoding for the environment.
# Consult https://click.palletsprojects.com/en/7.x/python3/ for mitigation steps.
#
# Edit 2019-10-08: whatsapp's wadebug uses "click.disable_unicode_literals_warning = True"
# Ref: https://github.com/WhatsApp/WADebug/blob/958ac37be804cc732ae514d4872b93d19d197a5c/wadebug/cli.py#L23
from ..utils import mysetlocale
mysetlocale()


import logging
logger = logging.getLogger('isitfit')

import click

from .. import isitfit_version


# For the --share-email "multiple options"
# https://click.palletsprojects.com/en/7.x/options/#multiple-options

@click.group(invoke_without_command=True)
@click.option('--debug', is_flag=True, help='Display more details to help with debugging')
@click.option('--optimize', is_flag=True, help='DEPRECATED: use "isitfit cost optimize" instead')
@click.option('--version', is_flag=True, help='DEPRECATED: use "isitfit version" instead')
@click.option('--share-email', multiple=True, help='Share result to email address')
@click.option('--skip-check-upgrade', is_flag=True, help='Skip step for checking for upgrade of isitfit')
@click.pass_context
def cli_core(ctx, debug, optimize, version, share_email, skip_check_upgrade):
    logLevel = logging.DEBUG if debug else logging.INFO
    ch = logging.StreamHandler()
    ch.setLevel(logLevel)
    logger.addHandler(ch)
    logger.setLevel(logLevel)

    if debug:
      logger.debug("Enabled debug level")
      logger.debug("-------------------")

    # After adding the separate command for "cost" (i.e. `isitfit cost analyze`)
    # putting a note here to notify user of new usage
    # Ideally, this code would be deprecated though
    if ctx.invoked_subcommand is None:
        # if still used without subcommands, notify user of new usage
        #from .cost import analyze as cost_analyze, optimize as cost_optimize
        #if optimize:
        #  ctx.invoke(cost_optimize, filter_tags=filter_tags, n=n)
        #else:
        #  ctx.invoke(cost_analyze, filter_tags=filter_tags)
        from click.exceptions import UsageError
        if optimize:
          raise UsageError("As of version 0.11, please use `isitfit cost optimize` instead of `isitfit --optimize`.")
        elif version:
          # ctx.invoke(cli_version)
          raise UsageError("As of version 0.11, please use `isitfit version` instead of `isitfit --version`.")
        else:
          raise UsageError("As of version 0.11, please use `isitfit cost analyze` instead of `isitfit` to calculate the cost-weighted utilization.")

    # make sure that context is a dict
    ctx.ensure_object(dict)

    # check if emailing requested
    if share_email is not None:
      max_n_recipients = 3
      if len(share_email) > max_n_recipients:
          from click.exceptions import BadParameter
          raise BadParameter("Maximum allowed number of email recipients is %i. Received %i"%(max_n_recipients, len(share_email)), param_hint="--share-email")

      ctx.obj['share_email'] = share_email

    # check if current version is out-of-date
    if not skip_check_upgrade:
      from ..utils import prompt_upgrade
      is_outdated = prompt_upgrade('isitfit', isitfit_version)
      ctx.obj['is_outdated'] = is_outdated



from .tags import tags as cli_tags
from .cost import cost as cli_cost
from .version import version as cli_version

cli_core.add_command(cli_version)
cli_core.add_command(cli_cost)
cli_core.add_command(cli_tags)

#-----------------------

if __name__ == '__main__':
  cli_core()



class Ec2Catalog:
  """
  download ec2 catalog: 2 columns: ec2 type, ec2 cost per hour
  """

  def handle_pre(self, context_pre):
    import requests
    from cachecontrol import CacheControl
    from cachecontrol.caches.file_cache import FileCache

    import logging
    logger = logging.getLogger('isitfit')
    logger.debug("Downloading ec2 catalog (cached to local file)")

    # based on URL = 'http://www.ec2instances.info/instances.json'
    # URL = 's3://...csv'
    # Edit 2019-09-10 use CDN link instead of direct gitlab link
    # URL = 'https://gitlab.com/autofitcloud/www.ec2instances.info-ec2op/raw/master/www.ec2instances.info/t3b_smaller_familyL2.json'
    URL = 'https://cdn.jsdelivr.net/gh/autofitcloud/www.ec2instances.info-ec2op/www.ec2instances.info/t3b_smaller_familyL2.json'

    # cached https://cachecontrol.readthedocs.io/en/latest/
    sess = requests.session()
    cached_sess = CacheControl(sess, cache=FileCache('/tmp/isitfit_ec2info.cache'))
    r = cached_sess.request('get', URL)

    # read catalog, copy from ec2op-cli/ec2op/optimizer/cwDailyMaxMaxCpu
    import json
    j = json.dumps(r.json(), indent=4, sort_keys=True)
    from pandas import read_json
    df = read_json(j, orient='split')
    
    # Edit 2019-09-13 no need to subsample the columns at this stage
    # df = df[['API Name', 'Linux On Demand cost']]

    df = df.rename(columns={'Linux On Demand cost': 'cost_hourly'})
    # df = df.set_index('API Name') # need to use merge, not index
    context_pre['df_cat'] = df

    return context_pre




class TestDotMan:

  def test_get_myuid_simple(self):
    from ..dotMan import DotMan
    uuid_val = DotMan().get_myuid()
    assert True

  # functional test
  def test_get_myuid_unavailable(self, monkeypatch):
    import tempfile
    from ..dotMan import DotMan
    dm = DotMan()
    with tempfile.TemporaryDirectory(prefix="isitfit-test-") as td:
      def mockreturn(): return td
      monkeypatch.setattr(dm, "get_dotisitfit", mockreturn)

      import os
      assert not os.path.isfile(os.path.join(td, "uid.txt"))
      uuid_val = dm.get_myuid()
      assert True

  # functional test
  def test_get_myuid_overwritten(self, monkeypatch):
    import tempfile
    from ..dotMan import DotMan
    dm = DotMan()
    with tempfile.TemporaryDirectory(prefix="isitfit-test-") as td:
      def mockreturn(): return td
      monkeypatch.setattr(dm, "get_dotisitfit", mockreturn)

      import os
      p3_uidtxt = os.path.join(td, "uid.txt")
      # overwrite
      with open(p3_uidtxt, 'w') as fh:
        fh.write("bla")

      # test core
      uuid_val = dm.get_myuid()
      assert uuid_val != "bla"
      assert len(uuid_val)==32






# Mostly a copy of test_iterator_redshift
# Need to convert the latter to class
# and inherit here to avoid code redundancy


from ....cost.redshift.iterator import Ec2Iterator

def test_init():
  rpi = Ec2Iterator()
  assert True # no exception


from moto import mock_ec2
@mock_ec2
def test_iterateCore_none(mocker):
  # mock the get regions part
  mockreturn = lambda service: ['us-east-1']
  mockee = 'boto3.session.Session.get_available_regions'
  mocker.patch(mockee, side_effect=mockreturn)

  # test
  rpi = Ec2Iterator()
  x = list(rpi.iterate_core())
  assert len(x) == 0


@mock_ec2
def test_iterateCore_exists(mocker):
  # mock the get regions part
  mockreturn = lambda service: ['us-east-1']
  mockee = 'boto3.session.Session.get_available_regions'
  mocker.patch(mockee, side_effect=mockreturn)

  # undo some region settings from before
  import boto3
  boto3.setup_default_session(region_name='us-east-1')

  # create mock redshift
  import boto3
  ec2_client = boto3.resource('ec2')
  ec2_client.create_instances(
    MinCount = 1,
    MaxCount = 1,
    InstanceType='t2.medium'
  )

  # test
  rpi = Ec2Iterator()
  rpi.region_include=['us-east-1']
  x = list(rpi.iterate_core())
  assert len(x) == 1


# cannot name function "test_iterator" because the filename is as such
# pytest .../test_iterator.py -k 'test_iterator' would run all tests, not just this one
@mock_ec2
def test_iteratorBuiltin(mocker):
  # create an instance
  import boto3
  ec2_client = boto3.resource('ec2')
  response_created = ec2_client.create_instances(
    MinCount = 1,
    MaxCount = 1,
    InstanceType='t2.medium'
  )
  response_created = response_created[0]

  # proceed
  import datetime as dt
  dt_now = dt.datetime.utcnow()

  # patch 1
  ex_iterateCore = [
    {'Region': 'us-east-1', 'InstanceId': response_created.instance_id}, # no creation time
    {'Region': 'us-east-1', 'InstanceId': response_created.instance_id, 'LaunchTime': dt_now}, # with creation time
  ]
  def mockreturn(*args, **kwargs):
    for x in ex_iterateCore:
      yield x

  mockee = 'isitfit.cost.redshift.iterator.BaseIterator.iterate_core'
  mocker.patch(mockee, side_effect=mockreturn)

  # patch 2
  #mockreturn = lambda *args, **kwargs: 1
  #mockee = 'isitfit.cost.redshift.iterator.BaseIterator.handle_cluster'
  #mocker.patch(mockee, side_effect=mockreturn)

  # patch 3
  ## import pandas as pd
  #mockreturn = lambda *args, **kwargs: 'a dataframe' #pd.DataFrame()
  #mockee = 'isitfit.cost.redshift.iterator.BaseIterator.handle_metric'
  #mocker.patch(mockee, side_effect=mockreturn)

  # test
  rpi = Ec2Iterator()
  x = list(rpi)
  assert len(x) == 1
  assert x[0][3] == response_created



def test_live_iterateCore():
  import os

  # reset all env vars from moto's mocks
  ev_l = ['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY', 'AWS_SECURITY_TOKEN', 'AWS_SESSION_TOKEN', 'AWS_DEFAULT_REGION']
  for ev_i in ev_l:
    if ev_i in os.environ.keys():
      del os.environ[ev_i]
 
  # set to profile
  os.environ["AWS_PROFILE"] = "shadi_shadi"

  from isitfit.cost.redshift.iterator import Ec2Iterator
  iterator = Ec2Iterator()
  expect_n = 4 # as of 2019-11-15

  # res = [x1 for x1 in iterator.iterate_core()]
  res = list(iterator.iterate_core())
  assert len(res) == expect_n

  # again with full iteration
  res = list(iterator)
  assert len(res) == expect_n

  # again with instance iterator
  # Note that this returns 3 entries instead of 4
  # because one instance doesn't have data
  # Edit 2019-11-18: moved cloudwatch part out of iterator,
  # and hence this returns 4 again
  res = list(iterator)
  assert len(res) == expect_n



import logging
logger = logging.getLogger('isitfit')


def cost_core(ra, rr, share_email, filter_region):
    """
    ra - Analyzer
    rr - Reporter
    """

    # data layer
    from .iterator import RedshiftPerformanceIterator
    ri = RedshiftPerformanceIterator(filter_region=filter_region)

    # pipeline
    from isitfit.cost.mainManager import MainManager
    from isitfit.cost.cacheManager import RedisPandas as RedisPandasCacheManager
    from .cloudwatchman import CloudwatchRedshift
    from isitfit.cost.ec2.ec2Common import Ec2Common

    mm = MainManager(None) # use None for click context for now FIXME
    cache_man = RedisPandasCacheManager()

    # manager of cloudwatch
    cwman = CloudwatchRedshift(cache_man)

    # common stuff
    ec2_common = Ec2Common()

    # update dict and return it
    # https://stackoverflow.com/a/1453013/4126114
    inject_email_in_context = lambda context_all: dict({'emailTo': share_email}, **context_all)
    inject_analyzer = lambda context_all: dict({'analyzer': ra}, **context_all)

    # setup pipeline
    mm.set_iterator(ri)
    mm.add_listener('pre', cache_man.handle_pre)
    mm.add_listener('ec2', cwman.per_ec2)
    mm.add_listener('ec2', ra.per_ec2)
    mm.add_listener('all', ec2_common.after_all) # just show IDs missing cloudwatch/cloudtrail
    mm.add_listener('all', ra.after_all)
    mm.add_listener('all', ra.calculate)
    mm.add_listener('all', inject_analyzer)
    mm.add_listener('all', rr.postprocess)
    mm.add_listener('all', rr.display)
    mm.add_listener('all', inject_email_in_context)
    mm.add_listener('all', rr.email)

    # kickoff
    mm.get_ifi()


def cost_analyze(share_email, filter_region):
  logger.info("Analyzing redshift clusters")

  from .analyzer import AnalyzerAnalyze
  from .reporter import ReporterAnalyze
  ra = AnalyzerAnalyze()
  rr = ReporterAnalyze()
  cost_core(ra, rr, share_email, filter_region)


def cost_optimize(filter_region):
  logger.info("Optimizing redshift clusters")

  from .analyzer import AnalyzerOptimize
  from .reporter import ReporterOptimize
  ra = AnalyzerOptimize()
  rr = ReporterOptimize()
  cost_core(ra, rr, None, filter_region)



import boto3
import pandas as pd
from tqdm import tqdm
import datetime as dt
import pytz

import logging
logger = logging.getLogger('isitfit')


from ..utils import SECONDS_IN_ONE_DAY, NoCloudwatchException, myreturn, NoCloudtrailException

MINUTES_IN_ONE_DAY = 60*24 # 1440
N_DAYS=90




from isitfit.cost.cacheManager import RedisPandas as RedisPandasCacheManager
class MainManager:
    def __init__(self, ctx):
        # set start/end dates
        dt_now_d=dt.datetime.now().replace(tzinfo=pytz.utc)
        self.StartTime=dt_now_d - dt.timedelta(days=N_DAYS)
        self.EndTime=dt_now_d
        logger.debug("Metrics start..end: %s .. %s"%(self.StartTime, self.EndTime))

        # listeners post ec2 data fetch and post all activities
        self.listeners = {'pre':[], 'ec2': [], 'all': []}

        # click context for errors
        self.ctx = ctx


    def set_iterator(self, ec2_it):
        # generic iterator (iterates over regions and items)
        self.ec2_it = ec2_it


    def add_listener(self, event, listener):
      if event not in self.listeners:
        from ..utils import IsitfitCliError
        err_msg = "Internal dev error: Event %s is not supported for listeners. Use: %s"%(event, ",".join(self.listeners.keys()))
        raise IsitfitCliError(err_msg, self.ctx)

      self.listeners[event].append(listener)


    def get_ifi(self):
        # 0th pass to count
        n_ec2_total = self.ec2_it.count()

        if n_ec2_total==0:
          return

        # context for pre listeners
        context_pre = {}
        context_pre['ec2_instances'] = self.ec2_it
        context_pre['region_include'] = self.ec2_it.get_regionInclude()
        context_pre['n_ec2_total'] = n_ec2_total
        context_pre['click_ctx'] = self.ctx


        # call listeners
        for l in self.listeners['pre']:
          context_pre = l(context_pre)
          if context_pre is None:
            raise Exception("Breaking the chain is not allowed in listener/pre")


        # iterate over all ec2 instances
        n_ec2_analysed = 0
        sum_capacity = 0
        sum_used = 0
        df_all = []
        ec2_noCloudwatch = []
        ec2_noCloudtrail = []

        # Edit 2019-11-12 use "initial=0" instead of "=1". Check more details in a similar note in "cloudtrail_ec2type.py"
        iter_wrap = tqdm(self.ec2_it, total=n_ec2_total, desc="Pass 2/2 through %s"%self.ec2_it.service_description, initial=0)
        for ec2_dict, ec2_id, ec2_launchtime, ec2_obj in iter_wrap:

          # context dict to be passed between listeners
          context_ec2 = {}
          context_ec2['mainManager'] = self
          if 'df_cat' in context_pre: context_ec2['df_cat'] = context_pre['df_cat'] # copy object between contexts
          context_ec2['ec2_dict'] = ec2_dict
          context_ec2['ec2_id'] = ec2_id
          context_ec2['ec2_launchtime'] = ec2_launchtime
          context_ec2['ec2_obj'] = ec2_obj

          try:
            n_ec2_analysed += 1

            # call listeners
            # Listener can return None to break out of loop,
            # i.e. to stop processing with other listeners
            for l in self.listeners['ec2']:
              context_ec2 = l(context_ec2)
              if context_ec2 is None: break

          except NoCloudwatchException:
            ec2_noCloudwatch.append(ec2_id)
          except NoCloudtrailException:
            ec2_noCloudtrail.append(ec2_id)

        # call listeners
        logger.info("... done")
        logger.info("")
        logger.info("")

        # set up context
        context_all = {}
        context_all['n_ec2_total'] = n_ec2_total
        context_all['mainManager'] = self
        context_all['n_ec2_analysed'] = n_ec2_analysed
        context_all['region_include'] = self.ec2_it.region_include
        if 'df_cat' in context_pre: context_all['df_cat'] = context_pre['df_cat'] # copy object between contexts

        # more
        context_all['ec2_noCloudwatch'] = ec2_noCloudwatch
        context_all['ec2_noCloudtrail'] = ec2_noCloudtrail
        context_all['click_ctx'] = self.ctx

        # call listeners
        for l in self.listeners['all']:
          context_all = l(context_all)
          if context_all is None:
            raise Exception("Breaking the chain is not allowed in listener/all: %s"%str(l))

        logger.info("")
        logger.info("")
        return





from ..utils import IsitfitCliError

import logging
logger = logging.getLogger('isitfit')

class TagsPush:
  """
  Class that will push a csv of tags to EC2
  Uses boto3's ResourceGroupsTaggingAPI
  for efficient "mass" tagging.
  https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/resourcegroupstaggingapi.html#ResourceGroupsTaggingAPI.Client.untag_resources
  https://docs.aws.amazon.com/resourcegroupstagging/latest/APIReference/API_UntagResources.html
  """

  def __init__(self, csv_fn, ctx):
    """
    csv_fn - filename of CSV file containing the tags
    ctx - click context object
    """
    self.csv_fn = csv_fn
    self.csv_df = None
    self.latest_df = None
    self.ctx = ctx

  def read_csv(self):
    import pandas as pd
    try:
      # read all fields as string
      self.csv_df = pd.read_csv(self.csv_fn, dtype=str)
    except pd.errors.EmptyDataError as e_info:
      raise IsitfitCliError("Error reading csv: %s"%str(e_info), self.ctx)

    if self.csv_df.shape[0]==0:
      raise IsitfitCliError("Tags csv file is empty", self.ctx)

    if 'instance_id' not in self.csv_df.columns:
      raise IsitfitCliError("Missing column instance_id", self.ctx)

    # sort by instance ID
    self.csv_df = self.csv_df.sort_values('instance_id', ascending=True)

    # fill na with ''
    self.csv_df = self.csv_df.fillna(value='')


  def validateTagsFile(self):
    if self.csv_df is None:
      raise IsitfitCliError("Internal dev error: Call TagsPush::read_csv before TagsPush::validateTagsFile", self.ctx)

    csv_dict = self.csv_df.to_dict(orient='records')
    from schema import Schema, Optional, SchemaError
    csv_schema = Schema([{
      'instance_id': str,
      'Name': str,
      Optional(str): str
    }])
    try:
      csv_schema.validate(csv_dict)
    except SchemaError as e:
      raise IsitfitCliError("CSV is not a tags file: %s"%str(e), self.ctx)

  def pullLatest(self):
    logger.info("Pulling latest tags for comparison")
    from .tagsDump import TagsDump
    td = TagsDump(self.ctx)
    td.fetch()
    td.suggest() # not really suggesting. Just dumping to csv
    self.latest_df = td.tags_df
    self.latest_df = self.latest_df.fillna(value='')

  def diffLatest(self):
    if self.latest_df is None:
      raise IsitfitCliError("Internal dev error: Call TagsPush::pullLatest before TagsPush::diffLatest", self.ctx)

    if self.csv_df is None:
      raise IsitfitCliError("Internal dev error: Call TagsPush::read_csv before TagsPush::diffLatest", self.ctx)

    # diff columns
    from .tagsCsvDiff import TagsCsvDiff
    td = TagsCsvDiff(self.latest_df, self.csv_df)
    td.noChanges()
    td.noNewInstances()
    td.getDiffCols()
    td.renamedTags()
    td.newTags()
    td.droppedTags()
    # print(td.migrations, td.old_minus_new, td.new_minus_old)
    td.anyRemaining()

    # get migrations
    import pandas as pd
    self.mig_df = pd.DataFrame(td.migrations, columns=['action', 'old', 'new'])
    logger.debug("")
    logger.debug("Tag migrations")
    if self.mig_df.shape[0]==0:
      logger.debug("None")
    else:
      logger.debug(self.mig_df)

    logger.debug("")

  def processPush(self, dryRun:bool):
    # max ec2 per call is 20
    # but just doing 1 at a time for now
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/resourcegroupstaggingapi.html#ResourceGroupsTaggingAPI.Client.tag_resources
    import boto3
    tagging_client = boto3.client('resourcegroupstaggingapi')
    ec2_resource = boto3.resource('ec2')
    account_id = boto3.client('sts').get_caller_identity()['Account']

    import json
    preproc = lambda x: x[sorted(list(x.columns))].set_index('instance_id')
    self.latest_df = preproc(self.latest_df)
    self.csv_df = preproc(self.csv_df)
    from tqdm import tqdm
    runType_prefix = "Dry run" if dryRun else "Live"
    for instance_id, row_new in tqdm(self.csv_df.iterrows(), total=self.csv_df.shape[0], desc="Tag CSV row (%s)"%runType_prefix, initial=1):
        row_old = self.latest_df.loc[instance_id]
        tags_new = row_new.to_dict()
        tags_old = row_old.to_dict()
        if tags_new==tags_old:
          logger.debug("Skipping %s since no changes"%instance_id)
          continue

        # keeping only changed keys
        keys_dotag = {}
        for k in tags_new:
          if not tags_new[k]:
            continue # empty tags are skipped

          if k not in tags_old:
            keys_dotag[k] = tags_new[k]
            continue

          if tags_new[k] != tags_old[k]:
            keys_dotag[k] = tags_new[k]
            continue

        # proceed with untagging
        keys_untag = []
        for k in tags_old:
          if not tags_old[k]:
            continue # empty tags are skipped

          if k not in tags_new:
            keys_untag.append(k)

        if not keys_dotag and not keys_untag:
          continue

        # if any of them set:
        instance_obj = ec2_resource.Instance(instance_id)
        instance_arn = 'arn:aws:ec2:%s:%s:instance/%s'%(instance_obj.placement['AvailabilityZone'][:-1], account_id, instance_id)

        if keys_dotag:
          logger.debug("[%s] Will tag %s with %s"%(runType_prefix, instance_id, json.dumps(keys_dotag)))
          if not dryRun:
            response = tagging_client.tag_resources(
              ResourceARNList=[instance_arn],
              Tags=keys_dotag
            )


        if keys_untag:
          logger.debug("[%s] Will untag %s with %s"%(runType_prefix, instance_id, json.dumps(keys_untag)))
          if not dryRun:
            response = tagging_client.untag_resources(
              ResourceARNList=[instance_arn],
              TagKeys=keys_untag
            )

    if dryRun:
      from termcolor import colored
      logger.info(colored("This was a dry run. Execute the same command again with `--not-dry-run` for actual tags push to aws ec2", "red"))



import logging
logger = logging.getLogger('isitfit')

import click

from ..utils import IsitfitCommand

@click.group(help="Explore EC2 tags", invoke_without_command=False)
def tags():
  pass



@tags.command(help="Generate new tags suggested by isitfit for each EC2 instance", cls=IsitfitCommand)
@click.option('--advanced', is_flag=True, help='Get advanced suggestions of tags. Requires login')
@click.pass_context
def suggest(ctx, advanced):
  # gather anonymous usage statistics
  from ..utils import ping_matomo, IsitfitCliError
  ping_matomo("/tags/suggest")

  tl = None
  if not advanced:
    from ..tags.tagsSuggestBasic import TagsSuggestBasic
    tl = TagsSuggestBasic(ctx)
  else:
    from ..tags.tagsSuggestAdvanced import TagsSuggestAdvanced
    tl = TagsSuggestAdvanced(ctx)

  tl.prepare()
  tl.fetch()
  tl.suggest()
  tl.display()


@tags.command(help="Dump existing EC2 tags in tabular form into a csv file", cls=IsitfitCommand)
@click.pass_context
def dump(ctx):
  # gather anonymous usage statistics
  from ..utils import ping_matomo, IsitfitCliError
  ping_matomo("/tags/dump")

  from ..tags.tagsDump import TagsDump
  tl = TagsDump(ctx)

  tl.fetch()
  tl.suggest() # not really suggesting. Just dumping to csv
  tl.display()



@tags.command(help="Push EC2 tags from csv file", cls=IsitfitCommand)
@click.argument('csv_filename') #, help='Path to CSV file holding tags to be pushed. Should match format from `isitfit tags dump`')
@click.option('--not-dry-run', is_flag=True, help='True for dry run (simulated push)')
@click.pass_context
def push(ctx, csv_filename, not_dry_run):
  # gather anonymous usage statistics
  from ..utils import ping_matomo, IsitfitCliError
  ping_matomo("/tags/push")

  from ..tags.tagsPush import TagsPush

  tp = TagsPush(csv_filename, ctx)

  tp.read_csv()
  tp.validateTagsFile()
  tp.pullLatest()
  tp.diffLatest()
  tp.processPush(not not_dry_run)




from .apiMan import ApiMan
import json
import logging
logger = logging.getLogger('isitfit')

from .utils import IsitfitCliError

class EmailMan:

  def __init__(self, dataType, dataVal, ctx):
    self.dataType = dataType
    self.dataVal = dataVal
    self.ctx = ctx
    self.api_man = ApiMan(tryAgainIn=1, ctx=ctx)

  def send(self, share_email):
    # get resources available
    self.api_man.register()

    # submit POST http request
    response_json, dt_now = self.api_man.request(
      method='post',
      relative_url='./share/email',
      payload_json={
        'dataType': self.dataType,
        'dataVal': self.dataVal,
        'share_email': share_email
      }
    )

    # check if error
    if response_json['isitfitapi_status']['code']=='Email verification in progress':
        raise IsitfitCliError(response_json['isitfitapi_status']['description'], self.ctx)

    if response_json['isitfitapi_status']['code']=='error':
        raise IsitfitCliError(response_json['isitfitapi_status']['description'], self.ctx)

    if response_json['isitfitapi_status']['code']!='ok':
        response_str = json.dumps(response_json)
        raise IsitfitCliError("Unsupported response from server: %s"%response_str, self.ctx)

    # validate schema
    from schema import SchemaError, Schema, Optional
    register_schema_2 = Schema({
        'from': str,
        Optional(str): object
      })
    try:
        register_schema_2.validate(response_json['isitfitapi_body'])
    except SchemaError as e:
        responseBody_str = json.dumps(response_json['isitfitapi_body'])
        err_msg = "Received response body: %s. Schema error: %s"%(responseBody_str, str(e))
        raise IsitfitCliError(err_msg, self.ctx)

    # otherwise proceed
    emailFrom = response_json['isitfitapi_body']['from']
    logger.info("Email sent from %s to: %s"%(emailFrom, ", ".join(share_email)))
    return






class TestMainManager:

  def test_addListener_failInvalid(self):
    from ...cost.mainManager import MainManager
    mm = MainManager(None)
    import pytest
    from ...utils import IsitfitCliError
    with pytest.raises(IsitfitCliError) as e:
      # raise exception        
      mm.add_listener('foo', lambda x: x)

    # check error message has "please upgrade" if ctx.obj.is_outdated = True
    # TODO






from ....cost.redshift.iterator import RedshiftPerformanceIterator

def test_init():
  rpi = RedshiftPerformanceIterator()
  assert True # no exception



from moto import mock_redshift
@mock_redshift
def test_iterateCore_none(mocker):
  # mock the get regions part
  mockreturn = lambda service: ['us-east-1']
  mockee = 'boto3.session.Session.get_available_regions'
  mocker.patch(mockee, side_effect=mockreturn)

  # test
  rpi = RedshiftPerformanceIterator()
  x = list(rpi.iterate_core())
  assert len(x) == 0


@mock_redshift
def test_iterateCore_exists(mocker):
  # mock the get regions part
  mockreturn = lambda service: ['us-east-1']
  mockee = 'boto3.session.Session.get_available_regions'
  mocker.patch(mockee, side_effect=mockreturn)

  # undo some region settings from before
  import boto3
  boto3.setup_default_session(region_name='us-east-1')

  # create mock redshift
  import boto3
  redshift_client = boto3.client('redshift')
  redshift_client.create_cluster(
    ClusterIdentifier='abc',
    NodeType='abc',
    MasterUsername='abc',
    MasterUserPassword='abc'
  )

  # test
  rpi = RedshiftPerformanceIterator()
  rpi.region_include = ['us-east-1']
  x = list(rpi.iterate_core())
  assert len(x) == 1


# cannot name function "test_iterator" because the filename is as such
# pytest .../test_iterator.py -k 'test_iterator' would run all tests, not just this one
def test_iteratorBuiltin(mocker):
  import datetime as dt
  dt_now = dt.datetime.utcnow()

  # patch 1
  ex_iterateCore = [
    {'ClusterIdentifier': 'abc'}, # no creation time
    {'ClusterIdentifier': 'abc', 'ClusterCreateTime': dt_now}, # with creation time
  ]
  mockreturn = lambda *args, **kwargs: ex_iterateCore
  mockee = 'isitfit.cost.redshift.iterator.RedshiftPerformanceIterator.iterate_core'
  mocker.patch(mockee, side_effect=mockreturn)

  # patch 2
  #mockreturn = lambda *args, **kwargs: 1
  #mockee = 'isitfit.cost.redshift.iterator.RedshiftPerformanceIterator.handle_cluster'
  #mocker.patch(mockee, side_effect=mockreturn)

  # patch 3
  ##import pandas as pd
  #mockreturn = lambda *args, **kwargs: 'a dataframe' #pd.DataFrame()
  #mockee = 'isitfit.cost.redshift.iterator.RedshiftPerformanceIterator.handle_metric'
  #mocker.patch(mockee, side_effect=mockreturn)

  # test
  rpi = RedshiftPerformanceIterator()
  x = list(rpi)
  assert len(x) == 1
  assert x[0][0] == ex_iterateCore[1]
  assert x[0][1] == 'abc' # 'a dataframe'




# imports
import datetime as dt
from ...utils import SECONDS_IN_ONE_DAY
import pandas as pd
import boto3

from isitfit.utils import raise_noCwExc, NoCloudwatchException, myreturn

import logging
logger = logging.getLogger('isitfit')


class CloudwatchBase:
  """
  Manager for cloudwatch
  """

  cloudwatch_namespace = None
  entry_keyId = None


  def __init__(self):
    self._initDates()


  def _initDates(self):
    # set start/end dates
    N_DAYS=90

    # FIXME? in mainManager, used pytz
    # dt_now_d=dt.datetime.now().replace(tzinfo=pytz.utc)
    dt_now_d = dt.datetime.utcnow()
    self.StartTime = dt_now_d - dt.timedelta(days=N_DAYS)
    self.EndTime = dt_now_d


  def _metric_get_statistics(self, metric):
    logger.debug("fetch cw")
    logger.debug(metric.dimensions)

    # util func
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudwatch.html#CloudWatch.Metric.get_statistics
    # https://docs.aws.amazon.com/redshift/latest/mgmt/metrics-listing.html
    response = metric.get_statistics(
        Dimensions=metric.dimensions,
        StartTime=self.StartTime,
        EndTime=self.EndTime,
        Period=SECONDS_IN_ONE_DAY,
        Statistics=['Minimum', 'Average', 'Maximum', 'SampleCount'],
        Unit = 'Percent'
    )
    return response


  def _metrics_filter(self, rc_id):
    if self.cloudwatch_namespace is None:
      raise Exception("Derived class should set cloudwatch_namespace")

    metrics_iterator = self.cloudwatch_resource.metrics.filter(
        Namespace = self.cloudwatch_namespace,
        MetricName = 'CPUUtilization',
        Dimensions=[
            {'Name': self.entry_keyId, 'Value': rc_id},
        ]
      )
    return metrics_iterator


  def handle_cluster(self, rc_id):

    #logger.debug("redshift cluster details")
    #logger.debug(rc_describe_entry)

    # remember that max for cluster = max of stats of all nodes
    logger.debug("Getting cloudwatch for cluster: %s"%(rc_id))
    metrics_iterator = self._metrics_filter(rc_id)
    for m_i in metrics_iterator:
        # skip node stats for now, and focus on cluster stats
        # i.e. dimensions only ClusterIdentifier, without the NodeID key
        if len(m_i.dimensions)>1:
          continue

        # exit the for loop and return this particular metric (cluster)
        return m_i

    # in case no cluster metrics found
    raise_noCwExc(rc_id)



  def handle_metric(self, m_i, rc_id, ClusterCreateTime):
    response_metric = self._metric_get_statistics(m_i)

    logger.debug("cw response_metric for %s"%rc_id)
    logger.debug(response_metric)

    if len(response_metric['Datapoints'])==0:
      raise_noCwExc(rc_id)

    # convert to dataframe
    df = pd.DataFrame(response_metric['Datapoints'])

    # edit 2019-09-13: no need to subsample columns
    # The initial goal was to drop the "Unit" column (which just said "Percent"),
    # but it's not such a big deal, and avoiding this subsampling simplifies the code a bit
    # df = df[['Timestamp', 'SampleCount', 'Average']]

    # sort and append in case of multiple metrics
    df = df.sort_values(['Timestamp'], ascending=True)

    # before returning, convert dateutil timezone to pytz
    # for https://github.com/pandas-dev/pandas/issues/25423
    # via https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.dt.tz_convert.html
    # Edit 2019-09-25 Instead of keeping the full timestamp, just truncate to date, especially that this is just daily data
    # df['Timestamp'] = df.Timestamp.dt.tz_convert(pytz.utc)
    df['Timestamp'] = df.Timestamp.dt.date

    # drop points "before create time" (bug in cloudwatch?)
    # Edit 2019-11-18 since this is daily data, and we don't really care about hours/minutes, just compare the y-m-d parts
    df = df[ df['Timestamp'] >= ClusterCreateTime.date() ]

    # print
    return df


  def _cloudwatch_metrics_boto3(self, region_name):
        """
        Easy-to-mock function since moto mock of cloudwatch is giving pagination error
        """
        boto3.setup_default_session(region_name = region_name)
        cloudwatch_resource = boto3.resource('cloudwatch')
        return cloudwatch_resource


  def handle_main(self, rc_describe_entry, rc_id, rc_created):
        logger.debug("Found cluster %s"%rc_id)

        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudwatch.html#metric
        self.cloudwatch_resource = self._cloudwatch_metrics_boto3(region_name = rc_describe_entry['Region'])

        # get metric
        m_i = self.handle_cluster(rc_id)

        # dataframe of CPU Utilization, max and min, over 90 days
        df = self.handle_metric(m_i, rc_id, rc_created)

        logger.debug("returning dataframe.head")
        logger.debug(df.head())

        return df


class CloudwatchCached(CloudwatchBase):
  def __init__(self, cache_man=None):
    """
    cache_man - from .cacheManager import RedisPandas as RedisPandasCacheManager

    """
    # manager of redis-pandas caching
    self.cache_man = cache_man
    super().__init__()


  def handle_main(self, rc_describe_entry, rc_id, rc_created):
        cache_key = "mainManager._cloudwatch_metrics/%s"%rc_id

        # in case of no cache
        if self.cache_man is None:
          df_fresh = super().handle_main(rc_describe_entry, rc_id, rc_created)
          return df_fresh

        # check cache first
        if self.cache_man.isReady():
          df_cache = self.cache_man.get(cache_key)
          if df_cache is not None:
            logger.debug("Found cloudwatch metrics in redis cache for %s, and data.shape[0] = %i"%(rc_id, df_cache.shape[0]))
            df_ret = myreturn(df_cache)
            if df_ret is None: raise_noCwExc(rc_id)
            if df_ret.shape[0]==0: raise_noCwExc(rc_id)
            return df_ret

        # if no cache, then download
        should_raise = False
        try:
          df_fresh = super().handle_main(rc_describe_entry, rc_id, rc_created)
        except NoCloudwatchException as e:
          df_fresh = pd.DataFrame()
          should_raise = True

        # if caching enabled, store it for later fetching
        # https://stackoverflow.com/a/57986261/4126114
        # Note that this even stores the result if it was "None" (meaning that no data was found)
        if self.cache_man.isReady():
          self.cache_man.set(cache_key, df_fresh)

        # check if should raise
        if should_raise:
          raise_noCwExc(rc_id)

        # done
        df_ret = myreturn(df_fresh)
        if df_ret is None: raise_noCwExc(rc_id)
        if df_ret.shape[0]==0: raise_noCwExc(rc_id)
        return df_ret



class CloudwatchRedshift(CloudwatchCached):
  cloudwatch_namespace = 'AWS/Redshift'
  entry_keyId = 'ClusterIdentifier'

  def per_ec2(self, context_ec2):
        """
        Raises NoCloudwatchException if no data found in cloudwatch
        """
        rc_describe_entry, rc_id, rc_created = context_ec2['ec2_dict'], context_ec2['ec2_id'], context_ec2['ec2_launchtime']
        df_single = self.handle_main(rc_describe_entry, rc_id, rc_created)
        context_ec2['df_single'] = df_single
        return context_ec2



class CloudwatchEc2(CloudwatchCached):
  cloudwatch_namespace = 'AWS/EC2'
  entry_keyId = 'InstanceId'

  def per_ec2(self, context_ec2):
        """
        Raises NoCloudwatchException if no data found in cloudwatch
        """
        ec2_obj = context_ec2['ec2_obj']
        df_cw3 = self.handle_main({'Region': ec2_obj.region_name}, ec2_obj.instance_id, ec2_obj.launch_time)
        context_ec2['df_metrics'] = df_cw3
        return context_ec2




import logging
logger = logging.getLogger('isitfit')

# URL of isitfit API
BASE_HOST = 'api.isitfit.io' # FIXME this should be in master
# BASE_HOST = 'api-dev.isitfit.io' # FIXME do not push this into master
BASE_PREFIX = 'v0'
BASE_URL = 'https://%s/%s/'%(BASE_HOST, BASE_PREFIX)

from .utils import IsitfitCliError
from schema import SchemaError, Schema, Optional


from aws_requests_auth.boto_utils import BotoAWSRequestsAuth
class MyBotoAWSRequestsAuth(BotoAWSRequestsAuth):
    def __init__(self, boto_session=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # over-write the credentials until a PR is made to integrate this upstream
        # https://github.com/DavidMuller/aws-requests-auth/blob/master/aws_requests_auth/boto_utils.py
        if boto_session is None: boto_session = boto3.session.Session()
        self._refreshable_credentials = boto_session.get_credentials()


class ApiMan:

  # number of seconds to wait between re-calls to register
  nsecs_wait = 30

  # counter of number of "register" calls
  call_n = 0 
  
  # class member holding register response
  r_register = None

  # class member shortcut for r_register['isitfitapi_body']
  r_body = None

  # number of times to call register before failing
  n_maxCalls = 3

  def __init__(self, tryAgainIn, ctx):
      self.tryAgainIn = tryAgainIn
      self.ctx = ctx

  def register(self):
      logger.debug("ApiMan::register")

      if self.r_register is not None:
          if self.r_register['isitfitapi_status']['code']=='ok':
              # already registered
              self.r_body = self.r_register['isitfitapi_body']
              return

      # counter
      self.call_n += 1

      # display head
      if self.call_n == 1:
          logger.info("Logging into server")
      else:
          logger.info("Registration attempt # %i."%(self.call_n))


      import boto3
      sts_client = boto3.client('sts')
      self.r_sts = sts_client.get_caller_identity()
      del self.r_sts['ResponseMetadata']
      # eg {'UserId': 'AIDA6F3WEM7AXY6Y4VWDC', 'Account': '974668457921', 'Arn': 'arn:aws:iam::974668457921:user/shadi'}

      # actual request
      self.r_register, dt_now = self.request(
          method='post',
          relative_url='./register',
          payload_json=self.r_sts,
          authenticated_user_path=False # since /register is the absolute path (without account/user)
        )

      # set shortcut to body only
      self.r_body = self.r_register['isitfitapi_body']

      # handle registration in progress
      if self.r_register['isitfitapi_status']['code']=='Registration in progress':
          # deal with "registration in progress"
          if self.call_n==1:
              # just continue and will check again later
              logger.info("Registration in progress")
          elif self.call_n >= self.n_maxCalls:
              raise IsitfitCliError("Registration is still not ready. Please try again in a few minutes, or file an issue at https://github.com/autofitcloud/isitfit/issues", self.ctx)
          else:
              logger.info("Registration not ready yet.")

          if self.call_n >= self.tryAgainIn:
              logger.info("Will check again in %i seconds"%(self.nsecs_wait))
              import time
              from tqdm import tqdm
              for i in tqdm(range(self.nsecs_wait)):
                time.sleep(1)

              self.register()
              return

          # stop here
          return


      # at this stage, registration was ok, so proceed
      if self.call_n==1:
          logger.info("Registration already done earlier")
      else:
          logger.info("Registration complete")


      # check schema
      register_schema_1 = Schema({
        's3_arn': str,
        's3_bucketName': str,
        's3_keyPrefix': str,
        'sqs_url': str,
        'role_arn': str,
        Optional(str): object
      })

      try:
        register_schema_1.validate(self.r_body)
      except SchemaError as e:
        import json
        logger.error("Received response: %s"%json.dumps(self.r_body))
        raise IsitfitCliError("Does not match expected schema: %s"%str(e), self.ctx)

      # show resources granted
      # print(self.r_register)
      logger.debug("AutofitCloud granted you access to the following AWS resources:")
      from tabulate import tabulate
      logger.debug(tabulate([(k, self.r_body[k]) for k in self.r_body.keys()]))
      logger.debug("Note that account number 974668457921 is AutofitCloud, the company behind isitfit.")
      logger.debug("For more info, visit https://autofitcloud.com/privacy")

      # get boto3 session using the assumed role
      # for further use of aws resources from AutofitCloud
      # eg API Gateway, S3, SQS
      sts_connection = boto3.client('sts')
      acct_b = sts_connection.assume_role(
                RoleArn=self.r_body['role_arn'],

                # TODO set external ID?
                #ExternalId=None,

                # this is a generic session name that shows up in the User ID field
                # on the server-side. If this gets modified, make sure to modify the
                # counter-part in isitfit-api
                RoleSessionName="cross_acct_isitfit"
      )
      self.boto3_session =  boto3.session.Session(
        aws_access_key_id=acct_b['Credentials']['AccessKeyId'],
        aws_secret_access_key=acct_b['Credentials']['SecretAccessKey'],
        aws_session_token=acct_b['Credentials']['SessionToken'],
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Queue.receive_messages
        # region matches with the serverless.yml region
        region_name='us-east-1'
      )

      # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Queue.receive_messages
      # region matches with the serverless.yml region
      sqs_res = self.boto3_session.resource('sqs') # no need for region since already in boto session # , region_name='us-east-1')
      self.sqs_q = sqs_res.Queue(self.r_body['sqs_url'])


  def request(self, method, relative_url, payload_json, authenticated_user_path=True):
      """
      Wrapper to the URL request
      method - post
      relative_url - eg ./tags/suggest
      payload_json - "json" field for request call
      authenticated_user_path - flag for self.register which can disable this as it doesn't have a account/user prefix in the URL
      """
      logger.debug("ApiMan::request")

      # relative URL to absolute
      # https://stackoverflow.com/a/8223955/4126114
      if authenticated_user_path:
        suffix_url='./%s/%s/%s'%(self.r_sts['Account'], self.r_sts['UserId'], relative_url)
      else:
        suffix_url = relative_url

      import urllib.parse
      absolute_url = urllib.parse.urljoin(BASE_URL, suffix_url)
      import json
      logger.debug("%s %s %s"%(method, absolute_url, json.dumps(payload_json)))

      # Either use a new boto session using the active AWS profile
      # or the boto session belonging to the assumed role
      import boto3
      boto_session=self.boto3_session if authenticated_user_path else boto3.session.Session()

      # prepare to use AWS Sigv4 with requests
      # https://stackoverflow.com/a/47252241/4126114
      #
      # use boto3 to collect credentials
      # https://github.com/DavidMuller/aws-requests-auth#using-boto-to-automatically-gather-aws-credentials
      #
      # original aws post (not clear)
      # https://docs.aws.amazon.com/general/latest/gr/sigv4-signed-request-examples.html
      #
      # clearer aws post
      # https://aws.amazon.com/premiumsupport/knowledge-center/iam-authentication-api-gateway/
      auth = MyBotoAWSRequestsAuth(aws_host=BASE_HOST,
                                    aws_region='us-east-1',
                                    aws_service='execute-api',
                                    boto_session=boto_session
                                    )

      # mark timestamp right before request (used in listen_sqs for dropping stale messages)
      import datetime as dt
      dt_now = dt.datetime.utcnow() #.strftime('%s')

      # make actual request
      import requests
      r1 = requests.request(method, absolute_url, json=payload_json, auth=auth)
      #logger.debug("python requests http request header:")
      #logger.debug(r1.request.headers)

      # https://stackoverflow.com/questions/18810777/how-do-i-read-a-response-from-python-requests
      import json
      r2 = json.loads(r1.text)

      # check AWS-generated errors
      if 'message' in r2:
        if r2['message']=='Internal server error':
          raise IsitfitCliError('Internal server error', self.ctx)
        else:
          # print(r2)
          raise IsitfitCliError('Serverside error #2: %s'%r2['message'], self.ctx)

      # validate schema of response
      register_schema_2 = Schema({
        'isitfitapi_status': {
            'code': str,
            'description': str,
        },
        'isitfitapi_body': {
          Optional(str): object
        }
      })
      try:
        register_schema_2.validate(r2)
      except SchemaError as e:
        import json
        logger.error("Received response: %s"%json.dumps(r2))
        raise IsitfitCliError("Does not match expected schema: %s"%str(e), self.ctx)

      # check for isitfit errors (in schema)
      if r2['isitfitapi_status']['code'] == 'error in schema':
        # print(r2)
        logger.debug("Detailed schema error message")
        logger.debug(r2['isitfitapi_status']['description'])
        raise IsitfitCliError("The data sent to the server by isitfit does not match the expected format.", self.ctx)

      # check for isitfit errors (general)
      elif r2['isitfitapi_status']['code'] == 'error':
        # print(r2)
        raise IsitfitCliError('Serverside error #1: %s'%r2['isitfitapi_status']['description'], self.ctx)

      elif r2['isitfitapi_status']['code']=='Registration in progress':
        # do nothing. This will be handled by self.register
        # Also, note that from isitfit-api, only the /register can lead to this code
        pass

      # check for isitfit unknown codes (i.e. maybe CLI is too old)
      elif r2['isitfitapi_status']['code'] != 'ok':
        # print(r2)
        msg = 'Unknown status code: %s. Description: %s'%(r2['isitfitapi_status']['code'], r2['isitfitapi_status']['description'])
        raise IsitfitCliError(msg, self.ctx)

      # if ok
      return r2, dt_now


  def listen_sqs(self, expected_type, dt_now):
    """
    expected_type - value of field "type" in the SQS messages to target
    dt_now - timestamp right before issuing the http that could have triggered the SQS messages
             Used to drop "stale" messages from earlier/canceled runs
    """
    # now listen on sqs

    # https://github.com/jegesh/python-sqs-listener/blob/master/sqs_listener/__init__.py#L123
    logger.info("Waiting for results")
    MAX_RETRIES = 5
    i_retry = 0
    import time
    n_secs = 5

    # loop
    while i_retry < MAX_RETRIES:
      i_retry += 1

      if i_retry == 1:
        time.sleep(1)
      else:
        #logger.info("Sleep %i seconds"%n_secs)
        time.sleep(n_secs)

      logger.debug("Check sqs messages (Retry %i/%i)"%(i_retry, MAX_RETRIES))
      messages = self.sqs_q.receive_messages(
        AttributeNames=['SentTimestamp'],
        QueueUrl=self.sqs_q.url,
        MaxNumberOfMessages=10
      )
      logger.debug("{} messages received".format(len(messages)))
      import datetime as dt
      import json
      for m in messages:
          sentTime_dt = None
          sentTime_str = "-"
          if m.attributes is not None:
            sentTime_dt = m.attributes['SentTimestamp']
            sentTime_dt = dt.datetime.utcfromtimestamp(int(sentTime_dt)/1000)
            sentTime_str = sentTime_dt.strftime("%Y/%m/%d %H:%M:%S")

          logger.debug("Message: %s: %s"%(sentTime_str, m.body))

          try:
            m.body_decoded = json.loads(m.body)
          except json.decoder.JSONDecodeError as e:
            logger.debug("(Invalid message with non-json body. Skipping)")
            continue

          if 'type' not in m.body_decoded:
            # print("FOOOOOOOOOO")
            logger.debug("(Message body missing key 'type'. Skipping)")
            continue

          if m.body_decoded['type'] != expected_type:
            logger.debug("(Message topic = %s != tags suggest. Skipping)")
            continue

          if (sentTime_dt < dt_now):
              logger.debug("Stale message (msg = %s < now = %s). Dropping and skipping"%(sentTime_dt, dt_now))
              m.delete()
              continue

          # all "tags suggest" messages are removed from the queue
          logger.debug("(Message is ok. Will process. Removing from queue)")
          m.delete()

          # process messages
          yield m

    # done
    yield None



from moto import mock_sts
import pytest
from ..apiMan import ApiMan
from ..utils import IsitfitCliError


@pytest.fixture(scope='function')
def MockRequestsRequest(mocker):
  def get_class(response):
    # set up
    def mockreturn(*args, **kwargs):
        return response
    mocker.patch('requests.request', side_effect=mockreturn)

    am = ApiMan(tryAgainIn=2, ctx=None)
    return am

  return get_class


class TestApiManRequest:
    @mock_sts
    def test_register_failSchemaL1(self, mocker, MockRequestsRequest):
        # since the MockApiManRequest patches the request function
        # and since the request function is
        response_val = {}
        class MockResponse:
          import json
          text = json.dumps(response_val)

        am = MockRequestsRequest(MockResponse)

        # trigger
        with pytest.raises(IsitfitCliError) as e:
            am.register()


    @mock_sts
    def test_register_failErrorGeneral(self, mocker, MockRequestsRequest):
        response_val = {'isitfitapi_status': {'code': 'error'}}
        class MockResponse:
          import json
          text = json.dumps(response_val)

        am = MockRequestsRequest(MockResponse)

        # trigger
        with pytest.raises(IsitfitCliError) as e:
            am.register()

#---------------------------

@pytest.fixture(scope='function')
def MockApiManRequest(mocker):
  def get_class(response):
    # set up
    def mockreturn(*args, **kwargs):
        return response, None
    mocker.patch('isitfit.apiMan.ApiMan.request', side_effect=mockreturn)

    am = ApiMan(tryAgainIn=2, ctx=None)
    return am

  return get_class


class TestApiManRegister:
    @mock_sts
    def test_register_failRegInProg(self, mocker, MockApiManRequest):
        response = {
                'isitfitapi_status': {'code': 'Registration in progress', 'description': 'foo'},
                'isitfitapi_body': {}
            }
        am = MockApiManRequest(response)
        am.nsecs_wait = 0

        # no exception, will not automatically try again
        am.call_n = 0
        am.tryAgainIn = 10
        am.register()

        # still no exception, will automatically try again till failing
        am.call_n = 1
        am.tryAgainIn = 2
        am.n_maxCalls = 5
        with pytest.raises(IsitfitCliError) as e:
          am.register()

        # triggers exception right away
        am.call_n = 2
        am.tryAgainIn = 2
        am.n_maxCalls = 3
        with pytest.raises(IsitfitCliError) as e:
            am.register()


    @mock_sts
    def test_register_failSchemaL2(self, mocker, MockApiManRequest):
        response = {
                'isitfitapi_status': {'code': 'ok', 'description': 'foo'},
                'isitfitapi_body': {
                }
            }
        am = MockApiManRequest(response)

        # exception
        with pytest.raises(IsitfitCliError) as e:
          am.register()


    @mock_sts
    def test_register_ok(self, mocker, MockApiManRequest):
        response = {
                'isitfitapi_status': {'code': 'ok', 'description': 'foo'},
                'isitfitapi_body': {
                    's3_arn': 'foo',
                    'sqs_url': 'foo',
                    'role_arn': '01234567890123456789',
                    's3_bucketName': 'foo',
                    's3_keyPrefix': 'foo',
                }
            }
        am = MockApiManRequest(response)

        # no exception
        am.register()
        assert True




import pytest
from ..utils import IsitfitCliError

@pytest.fixture(scope='function')
def MockEmailManFactory(mocker):
    def get_class(error_code, body):
        def mockreturn(*args, **kwargs):
            response = {'isitfitapi_status': {'code': error_code, 'description': 'foo'}, 'isitfitapi_body': body}
            return response, None

        mocker.patch('isitfit.apiMan.ApiMan.request', side_effect=mockreturn)
        mocker.patch('isitfit.apiMan.ApiMan.register', side_effect=lambda: None)

        # prepare
        from ..emailMan import EmailMan
        em = EmailMan(None, None, None)
        return em

    return get_class


def test_send_failVerificationInProgress(MockEmailManFactory):
    em = MockEmailManFactory('Email verification in progress', None)

    # trigger
    with pytest.raises(IsitfitCliError) as e:
      em.send([])


def test_send_failErrorGeneral(MockEmailManFactory):
    em = MockEmailManFactory('error', None)

    # trigger
    with pytest.raises(IsitfitCliError) as e:
      em.send([])


def test_send_failNotOk(MockEmailManFactory):
    em = MockEmailManFactory('hey', None)

    # trigger
    with pytest.raises(IsitfitCliError) as e:
      em.send([])


def test_send_failSchema(MockEmailManFactory):
    em = MockEmailManFactory('ok', {})

    # trigger
    with pytest.raises(IsitfitCliError) as e:
      em.send([])


def test_send_ok(MockEmailManFactory):
    em = MockEmailManFactory('ok', {'from': 'bla'})

    # no trigger
    em.send([])
    assert True # no exception



import logging
logger = logging.getLogger('isitfit')

import pandas as pd
from tabulate import tabulate
import tempfile
import csv
from collections import OrderedDict

# https://pypi.org/project/termcolor/
from termcolor import colored


def df2tabulate(df):
  return tabulate(df.set_index('instance_id'), headers='keys', tablefmt='psql')


def class2recommendedType(r):
  if r.classification_1 == 'Underused':
    # FIXME classification 2 will contain if it's a burstable workload or lambda-convertible
    # that would mean that the instance is downsizable twice, so maybe need to return r.type_smaller2x
    return r.type_smaller

  if r.classification_1=='Overused':
    return r.type_larger

  return None


def class2recommendedCost(r):
  if r.classification_1 == 'Underused':
    # FIXME add savings from the twice downsizing in class2recommendedType if it's a burstable workload or lambda-convertible,
    # then calculate the cost from lambda functions and add it as overhead here
    return r.cost_3m_smaller-r.cost_3m

  if r.classification_1=='Overused':
    return r.cost_3m_larger-r.cost_3m

  return None


def ec2obj_to_name(ec2_obj):
    if ec2_obj.tags is None:
      return None

    ec2_name = [x for x in ec2_obj.tags if x['Key']=='Name']
    if len(ec2_name)==0:
      return None

    return ec2_name[0]['Value']


class OptimizerListener:

  def __init__(self, n, thresholds = None):
    self.n = n

    if thresholds is None:
      thresholds = {
        'rightsize': {'idle': 3, 'low': 30, 'high': 70},
        'burst': {'low': 20, 'high': 80}
      }

    # iterate over all ec2 instances
    self.thresholds = thresholds
    self.ec2_classes = []

    # members that will contain the results of the optimization
    self.df_sort = None
    self.sum_val = None

    # for csv streaming
    self.csv_fn_intermediate = None
    self.csv_fh = None
    self.csv_writer = None

  
  def __exit__(self):
    self.csv_fh.close()


  def _xxx_to_classification(self, xxx_maxmax, xxx_maxavg, xxx_avgmax):
    # check if good to convert to burstable or lambda
    # i.e. daily data shows few large spikes
    thres = self.thresholds['burst']
    if xxx_maxmax >= thres['high'] and xxx_avgmax <= thres['low'] and xxx_maxavg <= thres['low']:
      return 'Underused', 'Burstable daily'

    # check rightsizing
    # i.e. no special spikes in daily data
    # FIXME: can check hourly data for higher precision here
    thres = self.thresholds['rightsize']
    if xxx_maxmax <= thres['idle']:
      return 'Idle', None
    elif xxx_maxmax <= thres['low']:
      return 'Underused', None
    elif xxx_maxmax >= thres['high'] and xxx_avgmax >= thres['high'] and xxx_maxavg >= thres['high']:
      return 'Overused', None
    elif xxx_maxmax >= thres['high'] and xxx_avgmax >= thres['high'] and xxx_maxavg <= thres['low']:
      return 'Underused', 'Burstable intraday'

    return 'Normal', None


  def _ec2df_to_classification(self, ec2_df, ddg_df):
    cpu_maxmax = ec2_df.Maximum.max()
    cpu_maxavg = ec2_df.Average.max()
    cpu_avgmax = ec2_df.Maximum.mean()
    cpu_c1, cpu_c2 = self._xxx_to_classification(cpu_maxmax, cpu_maxavg, cpu_avgmax)
    #print("ec2_df.{maxmax,avgmax,maxavg} = ", maxmax, avgmax, maxavg)

    if ddg_df is None:
      cpu_c2 = [cpu_c2, "No memory data"]
      cpu_c2 = [x for x in cpu_c2 if x is not None]
      cpu_c2 = ", ".join(cpu_c2)
      return cpu_c1, cpu_c2

    # continue with datadog data
    ram_maxmax = ddg_df.ram_used_max.max()
    ram_maxavg = ddg_df.ram_used_max.mean()
    ram_avgmax = ddg_df.ram_used_avg.max()
    ram_c1, ram_c2 = self._xxx_to_classification(ram_maxmax, ram_maxavg, ram_avgmax)

    # consolidate ram with cpu
    out_c2 = ["CPU + RAM checked",
              "CPU: %s"%cpu_c2 if cpu_c2 is not None else None,
              "RAM: %s"%ram_c2 if ram_c2 is not None else None ]
    out_c2 = ", ".join([x for x in out_c2 if x is not None])
    if cpu_c1=='Overused' or ram_c1=='Overused':
      return 'Overused', out_c2

    if cpu_c1=='Normal' or ram_c1=='Normal':
      return 'Normal', out_c2

    return 'Underused', out_c2


  def handle_pre(self, context_pre):
      # a csv file handle to which to stream results
      self.csv_fn_intermediate = tempfile.NamedTemporaryFile(prefix='isitfit-intermediate-', suffix='.csv', delete=False)
      logger.info(colored("Intermediate results will be streamed to %s"%self.csv_fn_intermediate.name, "cyan"))
      self.csv_fh = open(self.csv_fn_intermediate.name, 'w')
      self.csv_writer = csv.writer(self.csv_fh)

      # done
      return context_pre


  def per_ec2(self, context_ec2):
    # parse out context keys
    ec2_obj, ec2_df, mm, ddg_df = context_ec2['ec2_obj'], context_ec2['ec2_df'], context_ec2['mainManager'], context_ec2['ddg_df']

    #print(ec2_obj.instance_id)
    ec2_c1, ec2_c2 = self._ec2df_to_classification(ec2_df, ddg_df)

    ec2_name = ec2obj_to_name(ec2_obj)

    taglist = ec2_obj.tags

    # Reported in github issue 8: NoneType object is not iterable
    # https://github.com/autofitcloud/isitfit/issues/8
    if taglist is None:
      taglist = []

    if context_ec2['filter_tags'] is not None:
      # filter the tag list for only those containing the filter-tags string
      f_tn = context_ec2['filter_tags'].lower()
      # similar to the isitfit.mainManager.tagsContain function, but filtering the tags themselves
      taglist = [x for x in taglist if (f_tn in x['Key'].lower()) or (f_tn in x['Value'].lower())]

    taglist = ["%s = %s"%(x['Key'], x['Value']) for x in taglist]
    taglist = "\n".join(taglist)

    ec2_res = OrderedDict()
    ec2_res['region'] = ec2_obj.region_name
    ec2_res['instance_id'] = ec2_obj.instance_id
    ec2_res['instance_type'] = ec2_obj.instance_type
    ec2_res['classification_1'] = ec2_c1
    ec2_res['classification_2'] = ec2_c2
    ec2_res['tags'] = taglist

    # write csv header
    if len(self.ec2_classes)==0:
      self.csv_writer.writerow(['name'] + [k for k,v in ec2_res.items() if k!='tags'])# save header

    # save intermediate result to csv file
    # Try to stick to 1 row per instance
    # Drop the tags because they're too much to include
    csv_row = [ec2_name] + [v.replace("\n", ";") for k,v in ec2_res.items() if k!='tags']
    self.csv_writer.writerow(csv_row)

    # gathering results
    self.ec2_classes.append(ec2_res)

    # check if should return early
    if self.n!=0:
      sub_underused = [x for x in self.ec2_classes if x['classification_1']=='Underused']
      if len(sub_underused) >= self.n:
        # break early
        self.after_all(None, mm, None, None)
        import sys
        sys.exit(0)





from isitfit.cost.redshift.cli import cost_analyze, cost_optimize

import pytest

@pytest.mark.skip(reason="Need to figure out how to test this")
def test_costCore(mocker):
    mockee_list = [
      'isitfit.cost.redshift.iterator.RedshiftPerformanceIterator',
      'isitfit.cost.redshift.analyzer.AnalyzerAnalyze',
      'isitfit.cost.redshift.analyzer.AnalyzerOptimize',
      'isitfit.cost.redshift.reporter.ReporterAnalyze',
      'isitfit.cost.redshift.reporter.ReporterOptimize',
    ]
    for mockee_single in mockee_list:
      mocker.patch(mockee_single, autospec=True)

    # specific mocks
    # mocker.patch('isitfit.cost.redshift.iterator.RedshiftPerformanceIterator.count', side_effect=lambda: 1)

    # run and test
    cost_analyze(None)
    assert True # no exception

    cost_analyze([1])
    assert True # no exception

    cost_optimize()
    assert True # no exception



from ....cost.redshift.cloudwatchman import CloudwatchRedshift
import pytest
from isitfit.utils import NoCloudwatchException


def test_init():
  rpi = CloudwatchRedshift()
  assert True # no exception

# cannot use mock_cloudwatch
# yields error:
# botocore.exceptions.PaginationError: Error during pagination: The same next token was received twice
#from moto import mock_cloudwatch
#@mock_cloudwatch
def test_handleCluster_notFound(mocker):
  mockreturn = lambda *args, **kwargs: []
  mockee = 'isitfit.cost.redshift.cloudwatchman.CloudwatchRedshift._metrics_filter'
  mocker.patch(mockee, side_effect=mockreturn)

  rpi = CloudwatchRedshift()
  dummy_id = 'abc'

  with pytest.raises(NoCloudwatchException):
    m_i = rpi.handle_cluster(dummy_id)


def test_handleCluster_foundCluster(mocker):
  class MockMetricCluster:
    dimensions = [1]

  mockreturn = lambda *args, **kwargs: [MockMetricCluster]
  mockee = 'isitfit.cost.redshift.cloudwatchman.CloudwatchRedshift._metrics_filter'
  mocker.patch(mockee, side_effect=mockreturn)

  rpi = CloudwatchRedshift()
  dummy_id = 'abc'
  m_i = rpi.handle_cluster(dummy_id)
  assert m_i is not None


def test_handleCluster_foundMany(mocker):
  class MockMetricCluster:
    dimensions = [1]

  class MockMetricNode:
    dimensions = [1, 2]

  mockreturn = lambda *args, **kwargs: [MockMetricNode, MockMetricCluster]
  mockee = 'isitfit.cost.redshift.cloudwatchman.CloudwatchRedshift._metrics_filter'
  mocker.patch(mockee, side_effect=mockreturn)

  rpi = CloudwatchRedshift()
  dummy_id = 'abc'
  m_i = rpi.handle_cluster(dummy_id)
  assert m_i is not None


def test_handleMetric_empty(mocker):
  mockreturn = lambda *args, **kwargs: {'Datapoints': []}
  mockee = 'isitfit.cost.redshift.cloudwatchman.CloudwatchRedshift._metric_get_statistics'
  mocker.patch(mockee, side_effect=mockreturn)

  rpi = CloudwatchRedshift()
  with pytest.raises(NoCloudwatchException):
    df = rpi.handle_metric(None, None, None)



def test_handleMetric_notEmpty(mocker):
  import datetime as dt
  dt_now = dt.datetime.utcnow()

  ex_dp = [
    {'Timestamp': dt_now - dt.timedelta(seconds=1)},
    {'Timestamp': dt_now - dt.timedelta(seconds=2)},
    {'Timestamp': dt_now - dt.timedelta(seconds=3)}
  ]
  mockreturn = lambda *args, **kwargs: {'Datapoints': ex_dp}
  mockee = 'isitfit.cost.redshift.cloudwatchman.CloudwatchRedshift._metric_get_statistics'
  mocker.patch(mockee, side_effect=mockreturn)

  rpi = CloudwatchRedshift()
  df = rpi.handle_metric(None, None, dt_now)
  assert df is not None





# boto3 implementation of
# https://gist.github.com/shadiakiki1986/f6e676d1ab5800fcf7899b6a392ab821
# Docs
# https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudtrail.html#CloudTrail.Client.get_paginator
#
# Requirements: pip3 install boto3 tqdm pandas
# Run: python3 t2.py
#
# Edit 2019-09-13: copied file pull_cloudtrail_lookupEvents.py from git-remote-aws into isitfit so as to avoid confusion in download statistics
#----------------------------------------

# imports
import datetime as dt
from dateutil.relativedelta import relativedelta
import boto3
from tqdm import tqdm
import json

import logging
logger = logging.getLogger('isitfit')

#------------------------------
# utility to serialize date
#def json_serial(obj):
#    """JSON serializer for objects not serializable by default json code"""
#
#    if isinstance(obj, (dt.datetime, dt.date)):
#        return obj.isoformat()
#    raise TypeError ("Type %s not serializable" % type(obj))


#----------------------------------------
# iterate

# use jmespath like awscli
# https://stackoverflow.com/a/57018780/4126114
# Example
#   >>> mydata
#   {'foo': {'bar': [{'name': 'one'}, {'name': 'two'}]}}
#   >>> jmespath.search('foo.bar[?name==`one`]', mydata)
#   [{'name': 'one'}]
# import jmespath

#----------------------------------------
class Ec2Typechanges:
  
    def __init__(self, eventName):
      self.eventName = eventName

    # get paginator
    def getIterator(self, client):
        """
        eventName - eg 'ModifyInstanceAttribute'
        """
        # arguments to lookup-events command
        # From docs: "Currently the list can contain only one item"
        LookupAttributes=[
        #    {'AttributeKey': 'EventSource', 'AttributeValue': 'ec2.amazonaws.com'},
            {'AttributeKey': 'EventName', 'AttributeValue': self.eventName},
        ]

        # go back x time
        # https://stackoverflow.com/a/38795526/4126114
        # StartTime=dt.datetime.now() - relativedelta(years=1)
        # StartTime=dt.datetime.now() - relativedelta(days=90)
        PaginationConfig={
          'MaxResults': 3000
        }

        # client = boto3.client('cloudtrail')
        cp = client.get_paginator(operation_name="lookup_events")
        iterator = cp.paginate(
          LookupAttributes=LookupAttributes, 
          #StartTime=StartTime, 
          PaginationConfig=PaginationConfig
        )
        return iterator


    def handleIterator(self, iterator):
      i=1
      r_all = []
      for i, response in tqdm(enumerate(iterator), desc="Cloudtrail page %i"%i):
        #with open('t2.json','w') as fh:
        #  json.dump(response, fh, default=json_serial)

        # print(response.keys())
        for event in response['Events']:
          result = self._handleEventAny(event)
          if result is None: continue
          r_all.append(result)

      return r_all


    def _handleEventAny(self, event):
        if self.eventName == 'RunInstances':
          return self._handleEventRun(event)
          
        if self.eventName == 'ModifyInstanceAttribute':
          return self._handleEventModify(event)
          
        raise ValueError("Unsupported event name %s"%self.eventName)
      
      
    def _handleEventRun(self, event):
          # logger.debug("Cloudtrail event: %s"%json.dumps(event, default=json_serial))

          if 'Resources' not in event:
            logger.debug("No 'Resources' key in event. Skipping")
            return None # ignore this situation
        
          instanceId = [x for x in event['Resources'] if x['ResourceType']=='AWS::EC2::Instance']
          if len(instanceId)==0:
            logger.debug("No AWS EC2 instances in event. Skipping")
            return None # ignore this situation

          # proceed
          instanceId = instanceId[0]

          if 'ResourceName' not in instanceId:
            logger.debug("No ResourceName key in event. Skipping")
            return None # ignore this situation
          
          # proceed
          instanceId = instanceId['ResourceName']

          if 'CloudTrailEvent' not in event:
            logger.debug("No CloudTrailEvent key in event. Skipping")
            return None # ignore this situation

          ce_dict = json.loads(event['CloudTrailEvent'])

          if 'requestParameters' not in ce_dict:
            logger.debug("No requestParameters key in event['CloudTrailEvent']. Skipping")
            return None # ignore this situation

          if 'instanceType' not in ce_dict['requestParameters']:
            logger.debug("No instanceType key in event['CloudTrailEvent']['requestParameters']. Skipping")
            return None # ignore this situation

          newType = ce_dict['requestParameters']['instanceType']

          if 'EventTime' not in event:
            logger.debug("No EventTime key in event. Skipping")
            return None # ignore this situation

          ts_obj = event['EventTime']
          # ts_obj = dt.datetime.utcfromtimestamp(ts_int)
          # ts_str = ts_obj.strftime('%Y-%m-%d %H:%M:%S')

          result = {
            'EventTime': ts_obj,  # ts_str,
            'instanceId': instanceId,
            'instanceType': newType,
          }

          return result
          
          
    def _handleEventModify(self, event):
          if 'CloudTrailEvent' not in event:
            logger.debug("No CloudTrailEvent key in event. Skipping")
            return None # ignore this situation

          ce_dict = json.loads(event['CloudTrailEvent'])

          if 'requestParameters' not in ce_dict:
            logger.debug("No requestParameters key in event['CloudTrailEvent']. Skipping")
            return None # ignore this situation

          rp_dict = ce_dict['requestParameters']
          newType = None

          #newType = jmespath.search('instanceType', rp_dict)
          #if newType is None:
          #  newType = jmespath.search('attributeName==`instanceType`', rp_dict)

          if 'instanceType' in rp_dict:
            # logging.error(json.dumps(rp_dict))
            newType = rp_dict['instanceType']['value']

          if 'attribute' in rp_dict:
            if rp_dict['attribute']=='instanceType':
              newType = rp_dict['value']

          if newType is None:
            return None

          ts_obj = event['EventTime']
          # ts_obj = dt.datetime.utcfromtimestamp(ts_int)
          # ts_str = ts_obj.strftime('%Y-%m-%d %H:%M:%S')

          if 'instanceId' not in rp_dict:
            logger.debug("No instanceId key in requestParameters. Skipping")
            return None # ignore this situation

          result = {
            'EventTime': ts_obj, # ts_str,
            'instanceId': rp_dict['instanceId'],
            'instanceType': newType,
          }

          return result


class GeneralManager:
    def __init__(self, client):
        self.client = client
        
    def ec2_typeChanges(self):
        import pandas as pd
        from termcolor import colored
        import botocore
        import sys

        def run_iterator(man2_i, iterator_i):
          try:
            r_i = man2_i.handleIterator(iterator_i)
          except botocore.exceptions.ClientError as e:
            logger.error(colored("\n"+str(e), 'red'))
            sys.exit(1)

          return r_i

        man2_run = Ec2Typechanges(eventName='RunInstances')
        iterator_run = man2_run.getIterator(self.client)
        r_run = run_iterator(man2_run, iterator_run)
        
        man2_mod = Ec2Typechanges(eventName='ModifyInstanceAttribute')
        iterator_mod = man2_mod.getIterator(self.client)
        r_mod = run_iterator(man2_mod, iterator_mod)
       
        # split on instance ID and gather
        r_all = r_run + r_mod
        # logging.error(r_all)
        df = pd.DataFrame(r_all)

        if df.shape[0]==0:
          # early return
          return df

        df = df.set_index(["instanceId", "EventTime"]).sort_index()
        
        return df


if __name__=='__main__':
    client = boto3.client('cloudtrail')
    man1 = GeneralManager(client)
    df = man1.ec2_typeChanges()
    print("")
    print(df)



# mocker fixture becomes available after installing pytest-mock
# https://github.com/pytest-dev/pytest-mock
def test_ping_matomo(mocker):
  from ..utils import ping_matomo
  def mockreturn(url, json, timeout): return "foo"
  mocked_post = mocker.patch('isitfit.utils.requests.post', side_effect=mockreturn)
  ping_matomo("/test")

  # check that mocked object is called
  # https://github.com/pytest-dev/pytest-mock/commit/68868872195135bdb90d45a5cb0d609400943eae
  mocked_post.assert_called()



def test_isitfitCliError():
    import pytest
    from ..utils import IsitfitCliError

    class MockContext:
        obj = {'bar': 1}
        command = None

    ctx = MockContext()
    with pytest.raises(IsitfitCliError) as e:
        raise IsitfitCliError("foo", ctx)



from isitfit.cost.redshift.reporter import ReporterBase, ReporterAnalyze, ReporterOptimize

class TestReporterBase:
  def test_init(self):
    rb = ReporterBase()
    assert True


class TestReporterAnalyze:
  def test_postprocess(self):
    import datetime as dt
    dt_now = dt.datetime.utcnow()

    class MockMm:
      StartTime = dt_now
      EndTime = dt_now

    class MockAnalyzer:
      class MockIter:
        pass

      cwau_percent = 10
      rp_iter = MockIter
      n_rc_total = 0
      n_rc_analysed = 0
      regions_n = 1
      cost_billed = 1
      cost_used = 1

    rb = ReporterAnalyze()
    rb.postprocess({'analyzer': MockAnalyzer, 'mainManager': MockMm})
    assert rb.table is not None


  def test_display(self):
    import datetime as dt
    dt_now = dt.datetime.utcnow()

    class MockAnalyzer:
      class MockIter:
        rc_noData = []
      class MockCw:
        pass

      rp_iter = MockIter
      cwman = MockCw

    rb = ReporterAnalyze()
    rb.table = [
      {'color': '', 'label': 'bla', 'value': 'foo'}
    ]
    rb.analyzer = MockAnalyzer
    rb.display({})
    assert True # no exception


  def test_email(self, mocker):
    mockee = 'isitfit.emailMan.EmailMan'
    mocker.patch(mockee, autospec=True)
    rb = ReporterAnalyze()
    rb.table = []
    rb.email({'emailTo': [], 'click_ctx': None})
    assert True # no exception



class TestReporterOptimize:
  def test_postprocess(self):
    import pandas as pd
    class MockAnalyzer:
      analyze_df = pd.DataFrame([{'CpuMaxMax': 1, 'CpuMinMin': 1}])

    rb = ReporterOptimize()
    rb.postprocess({'analyzer': MockAnalyzer})
    assert True # no exception


  def test_display(self, mocker):
    mockee = 'isitfit.utils.display_df'
    mocker.patch(mockee, autospec=True)

    import pandas as pd
    class MockAnalyzer:
      analyze_df = pd.DataFrame([{'CpuMaxMax': 1, 'CpuMinMin': 1}])

    rb = ReporterOptimize()
    rb.csv_fn_final = 'bla.csv'
    rb.analyzer = MockAnalyzer
    rb.display({})
    assert True # no exception


  def test_email(self):
    import pytest
    with pytest.raises(Exception):
      rb = ReporterOptimize()
      rb.email({'emailTo': []})




# AWS_DEFAULT_REGION=us-east-2 python3 -m isitfit.cost.test_redshift
# Related
# https://docs.datadoghq.com/integrations/amazon_redshift/
# https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/redshift.html#Redshift.Paginator.DescribeClusters

from tqdm import tqdm
import pandas as pd
from isitfit.cost.mainManager import NoCloudwatchException

# redshift pricing as of 2019-11-12 in USD per hour, on-demand, ohio
# https://aws.amazon.com/redshift/pricing/
redshiftPricing_dict = {
  'dc2.large': 0.25,
  'dc2.8xlarge': 4.80,
  'ds2.xlarge': 0.85,
  'ds2.8xlarge': 6.80,
  'dc1.large': 0.25,
  'dc1.8xlarge': 4.80,
}
#redshiftPricing_df = [{'NodeType': k, 'Cost': v} for k, v in redshiftPricing_dict.items()]
#redshiftPricing_df = pd.DataFrame(redshiftPricing_df)
#print("redshift pricing")
#print(redshiftPricing_df)



class AnalyzerBase:


  def __init__(self):
    # define the list in the constructor because if I define it as a class member above,
    # then it gets reused between instantiations of derived classes
    self.analyze_list = []
    self.analyze_df = None

    self.n_rc_total = 0
    self.n_rc_analysed = 0


  def per_ec2(self, context_ec2):
      rc_describe_entry = context_ec2['ec2_dict']

      # for types not yet in pricing dictionary above
      rc_type = rc_describe_entry['NodeType']
      if rc_type not in redshiftPricing_dict.keys():
        raise NoCloudwatchException

      return context_ec2


  def after_all(self, context_all):
    # To be used by derived class *after* its own implementation

    # gather into a single dataframe
    self.analyze_df = pd.DataFrame(self.analyze_list)

    # update number of analyzed clusters
    self.n_rc_analysed = self.analyze_df.shape[0]

    if self.n_rc_analysed==0:
      from isitfit.utils import IsitfitCliError
      raise IsitfitCliError("No redshift clusters analyzed", context_all['click_ctx'])

    return context_all


  def calculate(self, context_all):
    raise Exception("To be implemented by derived class")



import datetime as dt
import pytz
import math
dt_now_d = dt.datetime.utcnow().replace(tzinfo=pytz.utc)

class AnalyzerAnalyze(AnalyzerBase):

  def per_ec2(self, context_ec2):
      # parent
      context_ec2 = super().per_ec2(context_ec2)

      # unpack
      rc_describe_entry, rc_id, rc_created = context_ec2['ec2_dict'], context_ec2['ec2_id'], context_ec2['ec2_launchtime']

      # get all performance dataframes, on the cluster-aggregated level
      df_single = context_ec2['df_single']

      rc_type = rc_describe_entry['NodeType']

      # append a n_hours field per day
      # and correct for number of hours on first and last day
      # Note that intermediate days are just 24 hours since Redshift cannot be stopped
      ymd_creation = rc_describe_entry['ClusterCreateTime'].strftime("%Y-%m-%d")
      ymd_today    = dt_now_d.strftime("%Y-%m-%d")

      hc_ref = dt_now_d if ymd_creation == ymd_today else rc_describe_entry['ClusterCreateTime'].replace(hour=23, minute=59)
      hours_creation = hc_ref - rc_describe_entry['ClusterCreateTime']
      hours_creation = math.ceil(hours_creation.seconds/60/60)
      hours_today = dt_now_d - dt_now_d.replace(hour=0, minute=0)
      hours_today = math.ceil(hours_today.seconds/60/60)

      def calc_nhours(ts):
        ts_str = ts.strftime("%Y-%m-%d")
        if ts_str == ymd_creation: return hours_creation
        if ts_str == ymd_today:    return hours_today
        return 24

      df_single['n_hours'] = df_single.Timestamp.apply(calc_nhours)

      # summarize into 1 row
      self.analyze_list.append({
        'ClusterIdentifier': rc_describe_entry['ClusterIdentifier'],
        'NodeType': rc_type,
        'NumberOfNodes': rc_describe_entry['NumberOfNodes'],
        'Region': rc_describe_entry['Region'],

        # TODO if cloudtrail is used to get the history of size changes of a redshift cluster,
        # the pricing field needs to change from a single number to a series of same dimension as df_single.Average
        # The same applies to the number of nodes field, which becomes a series too.
        # Finally, the "* df_single.shape[0]" becomes ".sum()"
        'CostUsed':   (df_single.Average / 100 * redshiftPricing_dict[rc_type] * df_single.n_hours * rc_describe_entry['NumberOfNodes']).sum(),
        'CostBilled': (1                       * redshiftPricing_dict[rc_type] * df_single.n_hours * rc_describe_entry['NumberOfNodes']).sum()
      })

      # done
      return context_ec2



  def calculate(self, context_all):
    # calculate cost-weighted utilization
    analyze_df = self.analyze_df
    self.cost_used   = analyze_df.CostUsed.fillna(value=0).sum()
    self.cost_billed = analyze_df.CostBilled.fillna(value=0).sum()
    self.regions_n = len(analyze_df.Region.unique())

    if self.cost_billed == 0:
      self.cwau_percent = 0
      return context_all

    self.cwau_percent = int(self.cost_used / self.cost_billed * 100)
    return context_all


class AnalyzerOptimize(AnalyzerBase):

  def per_ec2(self, context_ec2):
      """
      # get all performance dataframes, on the cluster-aggregated level
      """

      # parent
      context_ec2 = super().per_ec2(context_ec2)

      # unpack
      rc_describe_entry = context_ec2['ec2_dict']
      df_single = context_ec2['df_single']

      # summarize into maxmax, maxmin, minmax, minmin
      self.analyze_list.append({
        'Region': rc_describe_entry['Region'],
        'ClusterIdentifier': rc_describe_entry['ClusterIdentifier'],
        'NodeType': rc_describe_entry['NodeType'],
        'NumberOfNodes': rc_describe_entry['NumberOfNodes'],

        'CpuMaxMax': df_single.Maximum.max(),
        #'CpuMaxMin': df_single.Maximum.min(),
        #'CpuMinMax': df_single.Minimum.max(),
        'CpuMinMin': df_single.Minimum.min(),
      })

      # done
      return context_ec2



  def calculate(self, context_all):
    def classify_cluster_single(row):
        # classify
        if row.CpuMinMin > 70: return "Overused"
        if row.CpuMaxMax <  5: return "Idle"
        if row.CpuMaxMax < 30: return "Underused"
        return "Normal"

    # convert percentages to int since fractions are not very useful
    analyze_df = self.analyze_df
    analyze_df['classification'] = analyze_df.apply(classify_cluster_single, axis=1)
    return context_all



# imports
import datetime as dt
from ...utils import SECONDS_IN_ONE_DAY
import pandas as pd

import logging
logger = logging.getLogger('isitfit')

from isitfit.utils import IsitfitCliError


class BaseIterator:
  """
  Iterator design pattern
  Iterates over all CPU performance dataframes
  https://en.wikipedia.org/wiki/Iterator_pattern#Python
  """

  service_name = None
  service_description = None
  paginator_name = None
  paginator_entryJmespath = None
  paginator_exception = None
  entry_keyId = None
  entry_keyCreated = None


  def __init__(self, filter_region=None):
    # filter for certain region
    self.filter_region = filter_region

    # list of cluster ID's for which data is not available
    self.rc_noData = []

    # list of regions to skip
    self.region_include = []

    # in case of just_count=True, region_include is ignored since it is not yet populated
    # Set this flag to use region_include, eg if it is loaded from cache or if counting first pass is done
    self.regionInclude_ready = False

    # init cache
    self._initCache()

    # count of entries
    self.n_entry = None


  def get_regionInclude(self):
    """
    for the sake of mocking in the test tests/cost/redshift/test_cli.py:15
    """
    return self.region_include


  def _initCache(self):
    """
    # try to load region_include from cache
    """
    if self.filter_region is not None:
      self.region_include = [self.filter_region]
      self.regionInclude_ready = True
      return

    # need to use the profile name
    # because a profile could have ec2 in us-east-1
    # whereas another could have ec2 in us-west-1
    import boto3
    profile_name = boto3.session.Session().profile_name

    # cache filename and key to use
    from ...dotMan import DotMan
    import os
    self.cache_filename = 'iterator_cache-%s-%s.pkl'%(profile_name, self.service_name)
    self.cache_filename = os.path.join(DotMan().get_dotisitfit(), self.cache_filename)

    self.cache_key = 'iterator-region_include'

    # https://github.com/barisumog/simple_cache
    import simple_cache
    ri_cached = simple_cache.load_key(filename=self.cache_filename, key=self.cache_key)
    if ri_cached is not None:
      logger.debug("Loading regions containing EC2 from cache file")
      self.region_include = ri_cached
      self.regionInclude_ready = True


  def iterate_core(self, display_tqdm=False):
    fx_l = ['service_name', 'service_description', 'paginator_name', 'paginator_entryJmespath', 'paginator_exception', 'entry_keyId', 'entry_keyCreated']
    for fx_i in fx_l:
      # https://stackoverflow.com/a/9058315/4126114
      if fx_i not in self.__class__.__dict__.keys():
        raise Exception("Derived class should set %s"%fx_i)

    # iterate on regions
    import botocore
    import boto3
    import jmespath
    redshift_regions = boto3.Session().get_available_regions(self.service_name)
    # redshift_regions = ['us-west-2'] # FIXME

    if self.filter_region is not None:
      if self.filter_region not in redshift_regions:
        msg_err = "Invalid region specified: %s. Supported values: %s"
        msg_err = msg_err%(self.filter_region, ", ".join(redshift_regions))
        raise IsitfitCliError(msg_err, None) # passing None for click context

      # over-ride
      redshift_regions = [self.filter_region]

    # iterate
    region_iterator = redshift_regions
    if display_tqdm:
      from tqdm import tqdm
      region_iterator = tqdm(region_iterator, total = len(redshift_regions), desc="%s, counting in all regions"%self.service_description)

    for region_name in region_iterator:
      if self.regionInclude_ready and self.filter_region is None:
        if region_name not in self.region_include:
          # skip since already failed to use it
          continue

      logger.debug("Region %s"%region_name)
      boto3.setup_default_session(region_name = region_name)

      # boto3 clients
      # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/redshift.html#Redshift.Client.describe_logging_status
      redshift_client = boto3.client(self.service_name)

      # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudwatch.html#metric
      self.cloudwatch_resource = boto3.resource('cloudwatch')

      # iterate on redshift clusters
      paginator = redshift_client.get_paginator(self.paginator_name)
      rc_iterator = paginator.paginate()
      try:
        region_anyClusterFound = False
        for rc_describe_page in rc_iterator:
          rc_describe_entries = jmespath.search(self.paginator_entryJmespath, rc_describe_page)
          for rc_describe_entry in rc_describe_entries:
            region_anyClusterFound = True
            # add field for region
            rc_describe_entry['Region'] = region_name
            # yield
            yield rc_describe_entry

        if not self.regionInclude_ready and self.filter_region is None:
          if region_anyClusterFound:
            # only include if found clusters in this region
            self.region_include.append(region_name)

      except botocore.exceptions.ClientError as e:
        # Exception that means "no access to region"
        if e.response['Error']['Code']==self.paginator_exception:
          continue

        # all other exceptions raised
        raise e

    # before exiting, check if a count just completed, and mark region_include as usable
    if not self.regionInclude_ready and self.filter_region is None:
      self.regionInclude_ready = True

      # save to cache
      import simple_cache
      SECONDS_PER_HOUR = 60*60
      simple_cache.save_key(filename=self.cache_filename, key=self.cache_key, value=self.region_include, ttl=SECONDS_PER_HOUR)


  def count(self):
      # method 1
      # ec2_it = self.ec2_resource.instances.all()
      # return len(list(ec2_it))

    if self.n_entry is not None:
      return self.n_entry

    self.n_entry = len(list(self.iterate_core(True)))

    msg_count = "Found a total of %i %s in %i region(s) (other regions do not hold any %s)"
    logger.warning(msg_count%(self.n_entry, self.service_description, len(self.region_include), self.service_name))

    return self.n_entry


  def __iter__(self):
    for rc_describe_entry in self.iterate_core(False):
        #print("response, entry")
        #print(rc_describe_entry)

        # if not available yet (eg creating), still include analysis in case of past data
        #if rc_describe_entry['ClusterStatus'] != 'available':
        #    self.rc_noData.append(rc_id)
        #    continue

        if self.entry_keyId not in rc_describe_entry:
          # no ID, weird
          continue

        rc_id = rc_describe_entry[self.entry_keyId]

        if self.entry_keyCreated not in rc_describe_entry:
          # no creation time yet, maybe in process
          self.rc_noData.append(rc_id)
          continue

        rc_created = rc_describe_entry[self.entry_keyCreated]

        # None below is a placeholder for ec2_obj in case of ec2
        yield rc_describe_entry, rc_id, rc_created, None



class RedshiftPerformanceIterator(BaseIterator):
  service_name = 'redshift'
  service_description = 'Redshift clusters'
  paginator_name = 'describe_clusters'
  paginator_entryJmespath = 'Clusters[]'
  paginator_exception = 'InvalidClientTokenId'
  entry_keyId = 'ClusterIdentifier'
  entry_keyCreated = 'ClusterCreateTime'


class Ec2Iterator(BaseIterator):
  service_name = 'ec2'
  service_description = 'EC2 instances'
  paginator_name = 'describe_instances'
  # Notice that [] notation flattens the list of lists
  # http://jmespath.org/tutorial.html
  paginator_entryJmespath = 'Reservations[].Instances[]'
  paginator_exception = 'AuthFailure'
  entry_keyId = 'InstanceId'
  entry_keyCreated = 'LaunchTime'

  def __iter__(self):
    # over-ride the __iter__ to get the ec2 resource object for the current code (backwards compatibility)

    # method 1 for ec2
    # ec2_it = self.ec2_resource.instances.all()
    # return ec2_it

    # boto3 ec2 and cloudwatch data
    ec2_resource_all = {}
    import boto3

    # TODO cannot use directly use the iterator exposed in "ec2_it"
    # because it would return the dataframes from Cloudwatch,
    # whereas in the cloudwatch data fetch here, the data gets cached to redis.
    # Once the redshift.iterator can cache to redis, then the cloudwatch part here
    # can also be dropped, as well as using the "ec2_it" iterator directly
    # for ec2_dict in self.ec2_it:
    for ec2_dict, ec2_id, ec2_launchtime, _ in super().__iter__():
      if ec2_dict['Region'] not in ec2_resource_all.keys():
        boto3.setup_default_session(region_name = ec2_dict['Region'])
        ec2_resource_all[ec2_dict['Region']] = boto3.resource('ec2')

      ec2_resource_single = ec2_resource_all[ec2_dict['Region']]
      ec2_l = ec2_resource_single.instances.filter(InstanceIds=[ec2_dict['InstanceId']])
      ec2_l = list(ec2_l)
      if len(ec2_l)==0:
        continue # not found

      # yield first entry
      ec2_obj = ec2_l[0]
      ec2_obj.region_name = ec2_dict['Region']

      yield ec2_dict, ec2_id, ec2_launchtime, ec2_obj

